/* SAMHAIN file system integrity testing                                   */
/* Copyright (C) 1999 Rainer Wichmann                                      */
/*                                                                         */
/*  This program is free software; you can redistribute it                 */
/*  and/or modify                                                          */
/*  it under the terms of the GNU General Public License as                */
/*  published by                                                           */
/*  the Free Software Foundation; either version 2 of the License, or      */
/*  (at your option) any later version.                                    */
/*                                                                         */
/*  This program is distributed in the hope that it will be useful,        */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/*  GNU General Public License for more details.                           */
/*                                                                         */
/*  You should have received a copy of the GNU General Public License      */
/*  along with this program; if not, write to the Free Software            */
/*  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.              */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>

#include <errno.h>
#include <utime.h>

#include <sys/types.h>

#include "config_xor.h"

    

#ifdef HAVE_DIRENT_H
#include <dirent.h>
#define NAMLEN(dirent) strlen((dirent)->d_name)
#else
#define dirent direct
#define NAMLEN(dirent) (dirent)->d_namlen
#ifdef HAVE_SYS_NDIR_H
#include <sys/ndir.h>
#endif
#ifdef HAVE_SYS_DIR_H
#include <sys/dir.h>
#endif
#ifdef HAVE_NDIR_H
#include <ndir.h>
#endif
#endif

#include "samhain.h"

#if (defined (SH_WITH_CLIENT) || defined (SH_STANDALONE)) 

#include "sh_error.h"
#include "sh_utils.h"
#include "sh_unix.h"
#include "sh_files.h"
#include "sh_tiger.h"
#include "sh_hash.h"


typedef struct dir_struct {
  long    NumRegular;
  long    NumDirs;
  long    NumSymlinks;
  long    NumFifos;
  long    NumSockets;
  long    NumCDev;
  long    NumBDev;
  long    NumAll;
  long    TotalBytes;
  char    DirPath[PATH_MAX+2];
} dir_type;

typedef struct dirstack_entry {
  /*  char                  name[PATH_MAX+2]; */
  char                  * name;
  int                     class;
  struct dirstack_entry * next;
} dirstack_t;

static dirstack_t * dirList = NULL;

static dirstack_t * fileList = NULL;

static int        sh_files_fullpath  (char * testdir, char * d_name, 
				      char * statpath);
static int        sh_files_pushdir   (int class, char * str_s);
static int        sh_files_pushfile  (int class, char * str_s);
static int        sh_files_checkdir  (int class, char * dirName);
static ShFileType sh_files_filecheck (int class, char * dirName, 
				      char * fileName, long * total);

static long RecursionLevel = 0;
static long MaxRecursionLevel = 0;

/* set maximum recursion level
 */
int sh_files_setrecursion (char * flag_s)
{
  long flag = 0;

  if (flag_s != NULL) 
    flag = (int)(atof(flag_s));

  if (flag >= 0 && flag <= 99) {
    MaxRecursionLevel = flag;
  } else {
    sh_error_handle (SH_ERR_ERR, PLACE
		     NULL,
		     _("Invalid recursion level: %ld"),
		     flag);
    return (EXIT_FAILURE);
  }

  return (EXIT_SUCCESS);
}

void sh_files_chk ()
{
  dirstack_t * ptr;
  ShFileType   status;
  long         total;
  char       * base;
  char       * file;
  char       * tmp;
  
  ptr = fileList;

  while (ptr) {

    base = sh_util_basename (ptr->name);
    file = sh_util_filename (ptr->name);

    tmp = sh_util_safe_name (ptr->name);
    sh_error_handle (SH_ERR_INFO, PLACE
		     NULL,
		     _("Checking file %s"),
		     tmp);
    
    status = sh_files_filecheck (ptr->class, base, file, &total);
    
    if (status == SH_FILE_UNKNOWN) {
      sh_error_handle (ShDFLevel[SH_ERR_T_FILE], PLACE
		       NULL,
		       _("Failed checking file %s"),
		       tmp);
    }

    SH_FREE(tmp);
    SH_FREE(file);
    SH_FREE(base);

    ptr = ptr->next;
  }

  return;
}

int sh_files_delfilestack ()
{
  dirstack_t * old_ptr = fileList;

  while (old_ptr != NULL) {
    fileList = old_ptr->next;
    SH_FREE (old_ptr->name);
    SH_FREE (old_ptr);
    old_ptr = fileList;
  }
  return EXIT_SUCCESS;
}
  
int sh_files_deldirstack ()
{
  dirstack_t * old_ptr = dirList;

  while (old_ptr != NULL) {
    dirList = old_ptr->next;
    SH_FREE (old_ptr->name);
    SH_FREE (old_ptr);
    old_ptr = dirList;
  }
  return EXIT_SUCCESS;
}


int sh_files_pushfile_ro (char * str_s)
{
  return (sh_files_pushfile (SH_LEVEL_READONLY, str_s));
}

int sh_files_pushfile_log (char * str_s)
{
  return (sh_files_pushfile (SH_LEVEL_LOGFILES, str_s));
}

int sh_files_pushfile_glog (char * str_s)
{
  return (sh_files_pushfile (SH_LEVEL_LOGGROW, str_s));
}

int sh_files_pushfile_noig (char * str_s)
{
  return (sh_files_pushfile (SH_LEVEL_NOIGNORE, str_s));
}

int sh_files_pushfile_allig (char * str_s)
{
  return (sh_files_pushfile (SH_LEVEL_ALLIGNORE, str_s));
}

int sh_files_pushfile (int class, char * str_s)
{
  dirstack_t * new_item_ptr;
  char  * fileName;
  char  * tmp;
  int     len;

  if (sh.opts == 1) {
    sh_files_delfilestack ();
    sh_files_deldirstack ();
  }


  if (str_s == NULL) 
    return EXIT_FAILURE;

  len = strlen(str_s);

  if (len > PATH_MAX) {
    tmp = sh_util_safe_name (str_s);
    sh_error_handle (SH_ERR_ERR, PLACE
		     NULL,
		     _("File name too long: %s"),
		     tmp);
    SH_FREE(tmp);
    return EXIT_FAILURE;
  } else if (len < 1) {
    return EXIT_FAILURE;
  } else {
    fileName = SH_ALLOC(len+1);
    strlcpy(fileName, str_s, len+1);
  } 

  new_item_ptr = SH_ALLOC (sizeof(dirstack_t));

  new_item_ptr->name           = fileName;
  /*  strlcpy (new_item_ptr->name, fileName, PATH_MAX+1); */
  new_item_ptr->class          = class;
  new_item_ptr->next           = fileList;

  fileList = new_item_ptr;
  return EXIT_SUCCESS;
}

/* ------ directories ----- */

void sh_dirs_chk ()
{
  dirstack_t * ptr;
  int         status;
  
  ptr = dirList;

  while (ptr) {
    status = sh_files_checkdir (ptr->class, ptr->name);
    if (status == EXIT_FAILURE) {
      sh_error_handle (ShDFLevel[SH_ERR_T_DIR], PLACE
		       NULL,
		       _("Failed checking directory"));
    }
    ptr = ptr->next;
  }

  return;
}

int sh_files_pushdir_ro (char * str_s)
{
  return (sh_files_pushdir (SH_LEVEL_READONLY, str_s));
}

int sh_files_pushdir_log (char * str_s)
{
  return (sh_files_pushdir (SH_LEVEL_LOGFILES, str_s));
}

int sh_files_pushdir_glog (char * str_s)
{
  return (sh_files_pushdir (SH_LEVEL_LOGGROW, str_s));
}

int sh_files_pushdir_noig (char * str_s)
{
  return (sh_files_pushdir (SH_LEVEL_NOIGNORE, str_s));
}

int sh_files_pushdir_allig (char * str_s)
{
  return (sh_files_pushdir (SH_LEVEL_ALLIGNORE, str_s));
}


  
int sh_files_pushdir (int class, char * str_s)
{
  dirstack_t * new_item_ptr;
  char  * dirName;
  char  * tmp;
  int     len;

  if (sh.opts == 1) {
    sh_files_delfilestack ();
    sh_files_deldirstack ();
  }

  len = strlen(str_s);

  if (len > PATH_MAX) {
    tmp = sh_util_safe_name (str_s);
    sh_error_handle (SH_ERR_ERR, PLACE
		     NULL,
		     _("Directory name too long: %s"),
		     tmp);
    SH_FREE(tmp);
    return EXIT_FAILURE;
  } else if (len < 1) {
    return EXIT_FAILURE;
  } else {
    dirName = SH_ALLOC(len+1);
    strlcpy(dirName, str_s, len+1);
  } 

  new_item_ptr = SH_ALLOC (sizeof(dirstack_t));

  /*  strlcpy (new_item_ptr->name, dirName, PATH_MAX+1); */
  new_item_ptr->name           = dirName;
  new_item_ptr->class          = class;
  new_item_ptr->next           = dirList;

  dirList = new_item_ptr;
  return EXIT_SUCCESS;
}
  
  

int sh_files_checkdir (int class, char * dirName)
{
  DIR *           thisDir = NULL;
  struct dirent * thisEntry;
  int             status;
  dir_type        theDir;
  ShFileType      checkit;
  char            summary[10*16+1];
  char            dirHash[51];
  file_type       theFile;
  char          * tmpname;
  char          * tmpcat;

  tmpname = sh_util_safe_name (dirName);

  /* check for obscure name
   */
  sh_util_obscurename (ShDFLevel[SH_ERR_T_NAME], dirName);

  sh_error_handle (SH_ERR_INFO, PLACE
		   NULL, 
		   _("Entering directory %s"), 
		   tmpname);

  /* check input
   */
  if ( strlen(dirName) > PATH_MAX) {
    sh_error_handle (ShDFLevel[SH_ERR_T_DIR], PLACE
		     _("Path too long: %s"),
		     _("Path too long: %s (limit is %d)"), 
		     tmpname, (int) PATH_MAX);
    SH_FREE(tmpname);
    return EXIT_FAILURE;
  }
  if ( dirName[0] != '/') {
    sh_error_handle (ShDFLevel[SH_ERR_T_DIR], PLACE
		     NULL, 
		     _("Not an absolute path: %s"), 
		     tmpname);
    SH_FREE(tmpname);
    return EXIT_FAILURE;
  }
    
    
  /* stat the directory
   */
  strlcpy (theFile.fullpath, dirName, PATH_MAX+1);
  status = sh_unix_getinfo (ShDFLevel[SH_ERR_T_DIR], 
			    dirName,
			    &theFile);
  if (theFile.c_mode[0] != 'd') { 
    sh_error_handle (ShDFLevel[SH_ERR_T_DIR], PLACE
		     NULL,
		     _("Not a directory: %s"), 
		     tmpname);
    SH_FREE(tmpname); 
    return EXIT_FAILURE;
  }
  

  /* open directory for reading 
   */
  if ( (thisDir = opendir(dirName)) == NULL) {
    status = errno;
    sh_error_handle (ShDFLevel[SH_ERR_T_DIR], PLACE
		     _("Cannot read directory %s"),
		     _("Failed opendir() on %s: %s"),
		     tmpname, 
		     sh_error_message (status));
    SH_FREE(tmpname); 
    return EXIT_FAILURE;
  }

  theDir.NumRegular  = 0;
  theDir.NumDirs     = 0;
  theDir.NumSymlinks = 0;
  theDir.NumFifos    = 0;
  theDir.NumSockets  = 0;
  theDir.NumCDev     = 0;
  theDir.NumBDev     = 0;
  theDir.NumAll      = 0;
  theDir.TotalBytes  = 0;
  strlcpy (theDir.DirPath, dirName, PATH_MAX+1); 

  /* read
   */
  do {

    thisEntry = readdir (thisDir);

    if (thisEntry != NULL) {
      ++theDir.NumAll;

      if (strcmp (thisEntry->d_name, ".") == 0) 
	++theDir.NumDirs;

      else if (strcmp (thisEntry->d_name, "..") == 0)
	++theDir.NumDirs;

      else { 
	/* check the file 
	 */
	checkit = sh_files_filecheck (class, 
				      dirName, 
				      thisEntry->d_name,
				      &theDir.TotalBytes);
	if      (checkit == SH_FILE_REGULAR)   ++theDir.NumRegular;
	else if (checkit == SH_FILE_DIRECTORY) {
	  ++theDir.NumDirs;
	  if (RecursionLevel < MaxRecursionLevel) {
	    tmpcat = SH_ALLOC(PATH_MAX+2);
	    strlcpy(tmpcat, dirName,               PATH_MAX+2);
	    strlcat(tmpcat, "/",                   PATH_MAX+2);
	    strlcat(tmpcat, thisEntry->d_name,     PATH_MAX+2);
	    if (tmpcat != NULL) {
	      ++RecursionLevel;
	      sh_files_checkdir (class, tmpcat);
	      --RecursionLevel;
	    }
	    SH_FREE(tmpcat);
	  }
	}
	else if (checkit == SH_FILE_SYMLINK)   ++theDir.NumSymlinks;
	else if (checkit == SH_FILE_FIFO)      ++theDir.NumFifos;
	else if (checkit == SH_FILE_SOCKET)    ++theDir.NumSockets;
	else if (checkit == SH_FILE_CDEV)      ++theDir.NumCDev;
	else if (checkit == SH_FILE_BDEV)      ++theDir.NumBDev;
      }
    }
  } while (thisEntry != NULL);

  if (theDir.NumAll < 3) 
    sh_error_handle (SH_ERR_WARN, PLACE
		     NULL,
		     _("Found empty directory %s"),
		     tmpname);

  sh_error_handle (SH_ERR_INFO, PLACE
		   _(" d: %3ld, -: %3ld, l: %3ld, |: %3ld, s: %3ld, c: %3ld, b: %3ld"), 
		   _(" %ld directories, %ld files, %ld symlinks,\n   %ld fifos, %ld sockets, %ld charakter devices, %ld block devices"), 
		   theDir.NumDirs,
		   theDir.NumRegular,
		   theDir.NumSymlinks,
		   theDir.NumFifos,
		   theDir.NumSockets,
		   theDir.NumCDev,
		   theDir.NumBDev);

  sprintf(summary,                                    /* known to fit  */
	  _("%08lX%08lX%08lX%08lX%08lX%08lX%08lX%08lX"), 
	  theDir.NumDirs,
	  theDir.NumRegular,
	  theDir.NumSymlinks,
	  theDir.NumFifos,
	  theDir.NumSockets,
	  theDir.NumCDev,
	  theDir.NumBDev,
	  theDir.NumAll);
  
  /* compute checksum 
   *
   * 1 == init
   * 2 == check
   */
  if (sh.checkSum == SH_CHECK_INIT) {
    
    strlcpy(dirHash, sh_tiger_hash (summary, TIGER_DATA, strlen(summary)), 51);
    sh_hash_pushdata (&theFile, dirHash);
    sh_error_handle (SH_ERR_ALL, PLACE
		     NULL, _("  tiger checksum: %s"), dirHash);
    
  } else if (sh.checkSum == SH_CHECK_CHECK) {
    
    strlcpy(dirHash, sh_tiger_hash (summary, TIGER_DATA, strlen(summary)), 51);
    sh_hash_compdata (class, &theFile, dirHash);
    
  }

  sh_error_handle (SH_ERR_INFO, PLACE
		   NULL,
		   _("Leaving directory %s"), 
		   tmpname);
  closedir (thisDir);
  SH_FREE(tmpname);

  return EXIT_SUCCESS;
}

ShFileType sh_files_filecheck (int class, char * dirName, char * fileName,
			       long * total)
{
  char            fullpath[PATH_MAX+2];
  char            fileHash[51]; /* 3*16 + 2 blanks + '\0' */
  int             status;
  file_type       theFile;
  char          * tmpdir;
  char          * tmpname;
  struct utimbuf  utime_buf;

  if (EXIT_FAILURE == sh_util_obscurename (ShDFLevel[SH_ERR_T_NAME], 
				  fileName)) {
    tmpdir  = sh_util_safe_name (dirName);
    tmpname = sh_util_safe_name (fileName);
    sh_error_handle (ShDFLevel[SH_ERR_T_NAME], PLACE
		     NULL,
		     _("Bad filename: %s/%s"),
		     tmpdir, tmpname);
    SH_FREE(tmpname);
    SH_FREE(tmpdir);
  }    

  if (BAD == sh_files_fullpath (dirName, fileName, fullpath)) { 
    tmpdir  = sh_util_safe_name (dirName);
    tmpname = sh_util_safe_name (fileName);
    sh_error_handle (ShDFLevel[SH_ERR_T_FILE], PLACE
		     _("Filename too long: %s/%s"),
		     _("Filename too long: %s/%s (maximum %d)"),
		     tmpdir, tmpname, (int) PATH_MAX);
    SH_FREE(tmpname);
    SH_FREE(tmpdir);
    return SH_FILE_UNKNOWN; 
  } else {

    /* stat the file
     */
    strlcpy (theFile.fullpath, fullpath, PATH_MAX+1);
    status = sh_unix_getinfo (ShDFLevel[SH_ERR_T_FILE], 
			      fileName,
			      &theFile);

    if (status == EXIT_FAILURE) return SH_FILE_UNKNOWN;

    /* compute checksum 
     *
     * 1 == init
     * 2 == check
     */
    if (sh.checkSum == SH_CHECK_INIT && theFile.c_mode[0] == '-') {

      strlcpy(fileHash, 
	      sh_tiger_hash (fullpath, TIGER_FILE, 0),
	      51);
      sh_hash_pushdata (&theFile, fileHash);
      sh_error_handle (SH_ERR_ALL, PLACE
		     NULL, _("  tiger checksum: %s"), fileHash);

    } else if (sh.checkSum == SH_CHECK_CHECK 
	       && theFile.c_mode[0] == '-') {

      strlcpy(fileHash, 
	      sh_tiger_hash (fullpath, TIGER_FILE, 0), 51);
      sh_hash_compdata (class, &theFile, fileHash);

    }

    /* reset the access time 
     */
    if (class == SH_LEVEL_NOIGNORE)
      {
	utime_buf.actime   = (time_t) theFile.atime;
	utime_buf.modtime  = (time_t) theFile.mtime;
	sh_set_suid ();
	utime (fullpath, &utime_buf);
	sh_unset_suid ();
      }

    /* yes I know, the 'break' is not needed
     */
    switch (theFile.c_mode[0]) {
    case '-': return SH_FILE_REGULAR;   break;
    case 'l': return SH_FILE_SYMLINK;   break;
    case 'd': return SH_FILE_DIRECTORY; break;
    case 'c': return SH_FILE_CDEV;      break;
    case 'b': return SH_FILE_BDEV;      break;
    case '|': return SH_FILE_FIFO;      break;
    case 's': return SH_FILE_SOCKET;    break;
    default:  return SH_FILE_UNKNOWN;   break;
    }
	
  }

  /* notreached */
}


static int sh_files_fullpath (char * testdir, char * d_name, char * statpath)
{
  if (testdir != NULL) {
    if (strlen(testdir) > (PATH_MAX-2) ) return BAD;
    strlcpy(statpath, testdir,    PATH_MAX - 2);
  }
  if (d_name != NULL) {
    strlcat(statpath, "/",   PATH_MAX+1);
    if (strlen(d_name) >= ( PATH_MAX - strlen(statpath) ) ) 
      return BAD;
    strlcat(statpath, d_name,   PATH_MAX+1);
  }
  if (statpath == NULL) return BAD;
  return GOOD;
}

#endif
