/* SAMHAIN file system integrity testing                                   */
/* Copyright (C) 1999 Rainer Wichmann                                      */
/*                                                                         */
/*  This program is free software; you can redistribute it                 */
/*  and/or modify                                                          */
/*  it under the terms of the GNU General Public License as                */
/*  published by                                                           */
/*  the Free Software Foundation; either version 2 of the License, or      */
/*  (at your option) any later version.                                    */
/*                                                                         */
/*  This program is distributed in the hope that it will be useful,        */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/*  GNU General Public License for more details.                           */
/*                                                                         */
/*  You should have received a copy of the GNU General Public License      */
/*  along with this program; if not, write to the Free Software            */
/*  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.              */

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <sys/stat.h>
#include <unistd.h>

#include "config_xor.h"

#ifndef HAVE_LSTAT
#define lstat(a,b) stat(a,b)
#endif

#ifdef HAVE_MEMORY_H
#include <memory.h>
#endif

#if defined(SH_WITH_CLIENT) || defined(SH_STANDALONE)

#include "sh_hash.h"
#include "sh_utils.h"
#include "sh_error.h"
#include "sh_tiger.h"

/* MAX_PATH_STORE must be >= 50 
 */
#define MAX_PATH_STORE 254

typedef struct store_info {
  unsigned short   mark;
  char             fullpath[MAX_PATH_STORE+2];
  char             linkpath[MAX_PATH_STORE+2];
  unsigned int     mode;
  char             c_mode[11];
  long             owner;
  char             c_owner[OWNER_MAX+2];
  long             group;
  char             c_group[GROUP_MAX+2];
  unsigned long    ino;
  unsigned long    size;
  unsigned long    atime;
  unsigned long    mtime;
  unsigned long    ctime;
  char             checksum[KEY_LEN+1];
} sh_filestore_t;
  
typedef struct file_info {
  sh_filestore_t   theFile;
  int              visited;
  struct           file_info * next;
} sh_file_t;


/**********************************
 *
 * hash table functions
 *
 **********************************
 */

#include "sh_hash.h"

/* must fit an int */
#define TABSIZE 2048  

/* must fit an int */
#define REC_MAGIC 17  

static sh_file_t * tab[TABSIZE];

/* compute hash function
 */
static int hashfunc(char *s) {

  unsigned n = 0; 

  for ( ; *s; s++) 
    n = 31 * n + *s; 
  return n % TABSIZE; 
} 

/* search for files not visited, and check whether they exist
 */
static void hash_unvisited (sh_file_t *p, ShErrLevel level)
{
  struct stat buf;

  if (p->next != NULL)
    hash_unvisited (p->next, level);

  if (p->visited == BAD)
    {
      if (0 != lstat(p->theFile.fullpath, &buf))
	  sh_error_handle (level, PLACE
			   NULL, 
			   _("Not found, or access problem: %s."), 
			   p->theFile.fullpath);
    }

  p->visited = BAD;
  return;
}
void sh_hash_unvisited (ShErrLevel level)
{
  int i;

  for (i = 0; i < TABSIZE; ++i)
    {
      if (tab[i] != NULL) hash_unvisited (tab[i], level);
    }
  return;
}

/* delete hash array
 */
static void hash_kill (sh_file_t *p)
{

  if (p->next != NULL)
    hash_kill (p->next);

  SH_FREE(p);
  return;
}

/* get info out of hash array
 */
static sh_file_t * hashsearch (char * s) {

  sh_file_t * p;

  for (p = tab[hashfunc(s)]; p; p = p->next) 
    if (strcmp(s, p->theFile.fullpath) == 0) 
      return p; 
  return NULL;
} 

/* insert into hash array
 */
static void hashinsert (sh_file_t * s) {

  sh_file_t * p;

  int key = hashfunc(s->theFile.fullpath);

  if (tab[key] == NULL) {
    tab[key] = s;
    tab[key]->next = NULL;
    return;
  } else {
    p = tab[key];
    while (1) {
      if (p->next == NULL) {
	p->next = s;
	p->next->next = NULL;
	return;
      }
      p = p->next;
    }
  }
  /* notreached */
}


/* check functions
 */

static int IsInit = 0;

static void sh_hash_init ()
{
  sh_file_t * p;
  sh_filestore_t ft;
  int fd;
  int i;
  int count = 0;
  char * line;

  if (IsInit == 1) 
    return;

  for (i = 0; i < TABSIZE; ++i) tab[i] = NULL;

  if ( (fd = sh_unix_fileopen(DEFAULT_DATA_FILE, 'r', 
		       SH_ISDATA, SH_ERR_FATAL, YESPRIV)) == (-1))
    exit (EXIT_FAILURE);

  if (0 != strncmp(sh.filedata_hash, 
		   sh_tiger_hash (DEFAULT_DATA_FILE, 
				  TIGER_FILE, 0),
		   50)) {
    sh_error_handle (SH_ERR_FATAL, PLACE
		     NULL, 
		     _("Database: modified since program start."));
    exit (EXIT_FAILURE);
  }

  /* fast forward to start of data
   */
  line = SH_ALLOC(SH_BUFSIZE+1);
  while (1) {
    i =  sh_unix_getline (fd, line, SH_BUFSIZE);
    if (i < 0 ) {
      SH_FREE(line);
      sh_error_handle (SH_ERR_FATAL, PLACE
		       NULL, 
		       _("Database: no start marker found."));
      exit (EXIT_FAILURE);
    }
#ifdef SH_STEALTH
    if (0 == strncmp (line, N_("[SOF]"), 5)) 
#else
    if (0 == strncmp (line, _("[SOF]"),  5)) 
#endif
      break;
  }
  SH_FREE(line);

  while (1) {
    p = SH_ALLOC(sizeof(sh_file_t));
    ++count;
    if ( (i = sh_unix_fileread(fd, &ft, 
			       sizeof(sh_filestore_t))) <= 0 ) {
      SH_FREE(p);
      break;
    }
    if (ft.mark != REC_MAGIC)
      {
	SH_FREE(p);
	break;
      }
    memcpy( &(*p).theFile, &ft, sizeof(sh_filestore_t) );
    p->visited = BAD;

    hashinsert (p); 
  }

  sh_unix_fileclose (fd);
  IsInit = 1;
  return;
}
  
/* delete hash array
 */
void sh_hash_hashdelete ()
{
  int i;

  if (IsInit == 0) return;

  for (i = 0; i < TABSIZE; ++i) 
    if (tab[i] != NULL) hash_kill (tab[i]);

  IsInit = 0;
  return;
}

  
void sh_hash_pushdata (file_type * buf, char * fileHash)
{
  int fd;
  sh_filestore_t p;
  static int isfirst = 1;

  if ( (fd = sh_unix_fileopen(DEFAULT_DATA_FILE, 'w', 
		       SH_ISDATA, SH_ERR_FATAL, YESPRIV)) == (-1))
      return;

  if (strlen(buf->fullpath) <= MAX_PATH_STORE) {
    strlcpy(p.fullpath, buf->fullpath, MAX_PATH_STORE+1);  
  } else {
    strlcpy(p.fullpath, 
	    sh_tiger_hash (buf->fullpath, TIGER_DATA, strlen(buf->fullpath)), 
	    51);
  }

  if (buf->c_mode[0] == 'l') {  
    if (strlen(buf->linkpath) <= MAX_PATH_STORE) {
      strlcpy(p.linkpath, buf->linkpath, MAX_PATH_STORE+1);  
    } else {
      strlcpy(p.linkpath, 
	      sh_tiger_hash (buf->linkpath, TIGER_DATA, strlen(buf->linkpath)),
	      51);
    }
  }
  p.mark = REC_MAGIC;
  strlcpy(p.c_mode,   buf->c_mode,   11);
  strlcpy(p.c_group,  buf->c_group,  GROUP_MAX+1);
  strlcpy(p.c_owner,  buf->c_owner,  OWNER_MAX+1);
  strlcpy(p.checksum, fileHash,      KEY_LEN+1);

  p.mode  = buf->mode;
  p.ino   = buf->ino;
  p.size  = buf->size;
  p.mtime = buf->mtime;
  p.atime = buf->atime;
  p.ctime = buf->ctime;
  p.owner = buf->owner;
  p.group = buf->group;

  /* write the start marker 
   */
  if (isfirst == 1) {
#ifdef SH_STEALTH
    sh_unix_filewrite (fd, N_("\n[SOF]"), 6,
		       SH_ISLOG, SH_DATA_LINE);
#else
    sh_unix_filewrite (fd, _("\n[SOF]"),  6,
		       SH_ISLOG, SH_DATA_LINE);
#endif
    isfirst = 0;
  }
      
  sh_unix_filewrite (fd, &p, sizeof(sh_filestore_t), 
		     SH_ISDATA, SH_DATA_RAW);

  sh_unix_fileclose (fd);

  return;
}

int sh_hash_have_it (char * newname)
{
  sh_file_t * p;

  if (IsInit != 1) sh_hash_init();

  if (strlen(newname) <= MAX_PATH_STORE) {
    p = hashsearch(newname);
  } else {
    p = hashsearch ( sh_tiger_hash(newname, TIGER_DATA, strlen(newname)) );
  }
  if (p == NULL) return BAD;
  else return GOOD;
}
    
  
void sh_hash_compdata (int class,
		  file_type * theFile, char * fileHash)
{
  sh_file_t * p;
  char * tmp;
  char timstr1[80];
  char timstr2[80];
  char linkHash[KEY_LEN+1];
  int  maxcomp;

  if (IsInit != 1) sh_hash_init();


  /* --------  find the entry for the file ----------------       */

  if (strlen(theFile->fullpath) <= MAX_PATH_STORE) {
    p = hashsearch(theFile->fullpath);
  } else {
    p = hashsearch( sh_tiger_hash(theFile->fullpath, 
				  TIGER_DATA, 
				  strlen(theFile->fullpath))
		    );
  }
  if (p == NULL) {
    tmp = sh_util_safe_name(theFile->fullpath);
    sh_error_handle (ShDFLevel[class], PLACE
		  NULL, _("%s -- not found in database"),
		  tmp);
    SH_FREE(tmp);
    return;
  }
#ifdef MEM_DEBUG
  sh_error_handle (SH_ERR_ALL, PLACE
		   NULL, 
		   _("  tiger old: %s\n   tiger new: %s"), 
		   fileHash, 
		   p->theFile.checksum);
#endif


  if ( 
      (strncmp (fileHash, p->theFile.checksum, 50) != 0) && 
      ( class != SH_LEVEL_LOGFILES &&
	class != SH_LEVEL_LOGGROW  &&
	class != SH_LEVEL_ALLIGNORE 
      )
     ) {
    tmp = sh_util_safe_name(theFile->fullpath);
    sh_error_handle (ShDFLevel[class], PLACE
		     NULL, 
		     _("%s -- checksum changed %s -> %s"),
		     tmp,
		     p->theFile.checksum,
		     fileHash);
    SH_FREE(tmp);
  }

  if (p->theFile.c_mode[0] == 'l') {

    if (strlen(theFile->linkpath) <= MAX_PATH_STORE) {
      strlcpy(linkHash, 
	      sh_tiger_hash(theFile->linkpath, 
			    TIGER_DATA,
			    strlen(theFile->linkpath)), 
	      MAX_PATH_STORE+1);
      maxcomp = MAX_PATH_STORE;
    } else {
      strlcpy(linkHash, theFile->linkpath, 51);
      maxcomp = 50;
    }

    if ( 
	(strncmp (linkHash, p->theFile.linkpath, maxcomp) != 0) &&
	( class != SH_LEVEL_ALLIGNORE 
	  )
	) {
      tmp = sh_util_safe_name(theFile->linkpath);
      sh_error_handle (ShDFLevel[class], PLACE
		       NULL, 
		       _("%s -- link     changed (now %s)"),
		       tmp,
		       theFile->linkpath);
      SH_FREE(tmp);
    }
  }

    
  if ( (theFile->ino != p->theFile.ino ) &&
      ( class != SH_LEVEL_ALLIGNORE 
      )
     ) {
    tmp = sh_util_safe_name(theFile->fullpath);
    sh_error_handle (ShDFLevel[class], PLACE
		     NULL, 
		     _("%s -- inode    changed %ld -> %ld"),
		     tmp,
		     p->theFile.ino,
		     theFile->ino);
    SH_FREE(tmp);
  }
    
  if ( (theFile->owner != p->theFile.owner) &&
      ( class != SH_LEVEL_ALLIGNORE 
      )
     ) {
    tmp = sh_util_safe_name(theFile->fullpath);
    sh_error_handle (ShDFLevel[class], PLACE
		     NULL, 
		     _("%s -- owner    changed %s (%ld) -> %s (%ld)"),
		     tmp,
		     p->theFile.c_owner, p->theFile.owner,
		     theFile->c_owner, theFile->owner);
    SH_FREE(tmp);
  }

  if ( (theFile->group != p->theFile.group) &&
      ( class != SH_LEVEL_ALLIGNORE 
      )
    ) {
    tmp = sh_util_safe_name(theFile->fullpath);
    sh_error_handle (ShDFLevel[class], PLACE
		     NULL, 
		     _("%s -- group    changed %s (%ld) -> %s (%ld)"),
		     tmp,
		     p->theFile.c_group, p->theFile.group,
		     theFile->c_group, theFile->group);
    SH_FREE(tmp);
  }

  
  if ( (theFile->mtime != p->theFile.mtime) &&
      ( class != SH_LEVEL_LOGFILES &&
	class != SH_LEVEL_LOGGROW  &&
	class != SH_LEVEL_ALLIGNORE 
      )
    ) {
    strlcpy (timstr1, sh_unix_time (p->theFile.mtime), 80);
    strlcpy (timstr2, sh_unix_time (theFile->mtime), 80);
    tmp = sh_util_safe_name(theFile->fullpath);
    sh_error_handle (ShDFLevel[class], PLACE
		     NULL, 
		     _("%s -- mtime    changed %s -> %s"),
		     tmp,
		     timstr1,
		     timstr2);
    SH_FREE(tmp);
  }
  
  if ( (theFile->atime != p->theFile.atime) &&
      ( class == SH_LEVEL_NOIGNORE 
      )
    ) {
    strlcpy (timstr1, sh_unix_time (p->theFile.atime), 80);
    strlcpy (timstr2, sh_unix_time (theFile->atime), 80);
    tmp = sh_util_safe_name(theFile->fullpath);
    sh_error_handle (ShDFLevel[class], PLACE
		     NULL, 
		     _("%s -- atime    changed %s -> %s"),
		     tmp,
		     timstr1,
		     timstr2);
    SH_FREE(tmp);
  }
  
  /* Resetting the access time will set a new ctime. Thus, either we ignore
   * the access time or the ctime for NOIGNORE
   */
  if ( (theFile->ctime != p->theFile.ctime) &&
       ( class != SH_LEVEL_LOGFILES &&
	 class != SH_LEVEL_LOGGROW  &&
	 class != SH_LEVEL_ALLIGNORE &&
	 class != SH_LEVEL_NOIGNORE
	 )
       ) {
    strlcpy (timstr1, sh_unix_time (p->theFile.ctime), 80);
    strlcpy (timstr2, sh_unix_time (theFile->ctime), 80);
    tmp = sh_util_safe_name(theFile->fullpath);
    sh_error_handle (ShDFLevel[class], PLACE
   	 	     NULL, 
   		     _("%s -- ctime    changed %s -> %s"),
   		     tmp,
   		     timstr1,
   		     timstr2);
    SH_FREE(tmp);
  }
   
  
  if ( (theFile->size != p->theFile.size) &&
      ( class != SH_LEVEL_ALLIGNORE &&
        class != SH_LEVEL_LOGFILES  &&
	class != SH_LEVEL_LOGGROW   
      )
     ) {
    tmp = sh_util_safe_name(theFile->fullpath);
    sh_error_handle (ShDFLevel[class], PLACE
		     NULL, 
		     _("%s -- size     changed %ld -> %ld"),
		     tmp,
		     p->theFile.size,
		     theFile->size);
    SH_FREE(tmp);
  }
  
  if ( (theFile->size < p->theFile.size) &&
      ( class == SH_LEVEL_LOGGROW
      )
     ) {
    tmp = sh_util_safe_name(theFile->fullpath);
    sh_error_handle (ShDFLevel[class], PLACE
		     NULL, 
		     _("%s -- size    shrinked %ld -> %ld"),
		     tmp,
		     p->theFile.size,
		     theFile->size);
    SH_FREE(tmp);
  }

  p->visited = GOOD;

  return;
}
  


/* if defined(SH_WITH_CLIENT) || defined(SH_STANDALONE) */

#endif
