/* SAMHAIN file system integrity testing                                   */
/* Copyright (C) 1999 Rainer Wichmann                                      */
/*                                                                         */
/*  This program is free software; you can redistribute it                 */
/*  and/or modify                                                          */
/*  it under the terms of the GNU General Public License as                */
/*  published by                                                           */
/*  the Free Software Foundation; either version 2 of the License, or      */
/*  (at your option) any later version.                                    */
/*                                                                         */
/*  This program is distributed in the hope that it will be useful,        */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/*  GNU General Public License for more details.                           */
/*                                                                         */
/*  You should have received a copy of the GNU General Public License      */
/*  along with this program; if not, write to the Free Software            */
/*  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.              */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <pwd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <signal.h>
#include <setjmp.h>


#include "config_xor.h"

#ifdef HAVE_MEMORY_H
#include <memory.h>
#endif

#include "samhain.h"
#include "sh_error.h"
#include "sh_unix.h"
#include "sh_tiger.h"
#include "sh_mail.h"
#include "sh_utils.h"
#include "sh_fifo.h"

static int failedMail = GOOD;

/*******************************
typedef struct mail_msg {
  char   * msg;
  struct mail_msg * next;
} mail_msg_t;
static mail_msg_t * mailStack = NULL;
********************************/

/*********************************************
 *  utility function for verifying mails
 *********************************************/

int sh_mail_sigverify (char * s)
{
  int    fd;
  int    i;
  char * buf;
  char * bufc;
  char   key[81];
  char   number[81];
  long   numsig;
  char   key2[KEY_MAX+1];
  char   key3[KEY_MAX+1];

  /* sh_unix_droppriv (); */

  sh_error_logoff();

  if (s == NULL || strlen(s) >= PATH_MAX) 
    exit (EXIT_FAILURE);

  /* open the file, then check it 
   */
  if ( (fd = sh_unix_fileopen (s, 'r', 
			       SH_ISFILE, SH_ERR_FATAL, NOPRIV)) == (-1))
    {
      fprintf(stderr, _("Could not open file %s\n"), s);
      _exit (EXIT_FAILURE);
    }

  buf     = SH_ALLOC( (unsigned long)(SH_BUFSIZE+1));
  bufc    = SH_ALLOC( (unsigned long)(SH_MAXBUF+1));

  while (1)
    {
      buf[0]  = '\0';
      bufc[0] = '\0';

      /* find start of next message
       */
      while (0 != strncmp(buf,   _("<--- MESSAGE   ---->"),
			  sizeof("<--- MESSAGE   ---->")-1)) 
	{
	  (void) sh_unix_getline (fd, buf, SH_BUFSIZE);
	  if (buf[0] == '\0')  /* end of mailbox reached, exit */
	    {
	      fflush(stdout);
	      _exit (EXIT_SUCCESS);
	    }
	}
      
      /* read message, compress into bufc
       */
      while (1)
	{
	  (void) sh_unix_getline (fd, buf, SH_BUFSIZE);
	  if (0 == strncmp(buf,   _("<--- SIGNATURE ---->"),
			   sizeof("<--- SIGNATURE ---->")-1))
	    break;
	  if (buf[0] == '\0') 
	    exit (EXIT_FAILURE);
	  sh_util_compress(bufc, buf, SH_MAXBUF-KEY_MAX);
	}
      
      /* get signature and number 
       */
      (void) sh_unix_getline (fd, key,    sizeof(key));
      key[KEY_MAX] = '\0';
      (void) sh_unix_getline (fd, number, sizeof(number));
      number[63]   = '\0';
      numsig = atol (number);
      
      fprintf(stderr, _("Message %s - Signature %s\n"), number, key);

      if (numsig == 0)
	{
	  sh_util_encode(key, bufc, 1);
	  strlcpy(key3, key, KEY_MAX+1);
	  fprintf (stderr, _("(unchecked)\n"));
	}
      else
	{
	  /* iterate key
	   */
	  strlcpy(key2, key3, KEY_MAX+1); 
	  for (i = 0; i < numsig; ++i) 
	    {
	      strlcpy (key2, 
		       sh_tiger_hash (key2, TIGER_DATA, KEY_MAX), 
		       KEY_MAX+1);
	    }
	  
	  strlcat (bufc, key2, SH_MAXBUF+1);

	  if (strncmp (key, 
		       sh_tiger_hash (bufc, TIGER_DATA, strlen (bufc)),
		       50) != 0) {
	    fprintf (stderr, _("(FAILED)\n"));
	  } else { 
	    fprintf (stderr, _("(passed)\n"));
	  }
	}

    } /* end scan mailbox */

  /* notreached */
}

int sh_mail_setaddress (char * address)
{
  unsigned long  numcop;

  if (address != NULL) 
    {
      numcop = strlcpy (sh.address, address, 256+1);
      if (numcop >= 257) 
	{
	  fprintf(stderr, 
		  _("ERROR : Mail address too long (max=256): %s"), 
		  address);
	  return BAD;
	}
      return GOOD;
    }
  return BAD;
}

int sh_mail_setNum (char * str)
{
  int i = atoi (str);

  if (i >= 0 && i < SH_FIFO_MAX) {
    sh.mailNum.alarm_interval = i;
    return GOOD;
  } else {
    sh_error_handle (SH_ERR_ERR, PLACE
		     NULL,
		     _("Invalid argument (%s) for max pending "\
		       "messages on mail queue (0 to %d)"),
		     str, SH_FIFO_MAX);
    return BAD;
  }
}


static SH_FIFO * fifo_mail = NULL;

void sh_mail_emptystack ()
{
  char * msg;
  int    len;

  if (fifo_mail == NULL)
    return;

  while (NULL != (msg = pop_list(fifo_mail)))
    {
      len = strlen(msg);
      memset(msg, 0, len);
      SH_FREE(msg);
    }

  return;
}

void sh_mail_pushstack (char * msg)
{
  if (msg == NULL || failedMail == BAD) 
    return;

  if (fifo_mail == NULL)
    {
      fifo_mail = SH_ALLOC(sizeof(SH_FIFO));
      fifo_init(fifo_mail);
    }

  push_list (fifo_mail, msg);
  ++sh.mailNum.alarm_last;

  if (sh.mailNum.alarm_last >= sh.mailNum.alarm_interval)
    sh_mail_msg (sh.address, NULL);

  return;
}

/* the mailer
 */
static int sh_mail_end_conn (FILE * connfile);
static FILE * sh_mail_start_conn (char * address);


#if defined(HAVE_MLOCK) && !defined(HAVE_BROKEN_MLOCK)
#include <sys/mman.h>
#endif

int sh_mail_msg (char * address, char * message)
{
    char         subject[32+32+SH_MINIBUF+2+3];
    char       * mailMsg;
    int          status;
    char       * bufcompress;
    static int   failcount = 0;
    static int   isfirst   = 1;
    static int   mailcount = 0;
    FILE       * connfile  = NULL;

    struct  sigaction  old_act;
    struct  sigaction  new_act;


    /* return if we cannot mail
     */
    if (failedMail == BAD) 
      return EXIT_FAILURE;

    new_act.sa_handler = SIG_IGN;

    /* catch (ignore) 'broken pipe'
     */
    sigaction (SIGPIPE, &new_act, &old_act);

    /* ---------- Connect ---------------------------------------- */

    if (NULL == (connfile = sh_mail_start_conn (address)) )
      {
	sh_error_handle (SH_ERR_FATAL, PLACE
			 NULL,
			 _("Mail error - could not open connection."));
	if (connfile != NULL) fclose (connfile);
	sigaction (SIGPIPE, &old_act, NULL);

	sh_mail_pushstack (message);  /* save for later */
	++failcount;
	if (failcount > SH_MAX_FAIL)
	  {
	    sh_mail_emptystack();
	    failedMail = BAD;
	  }

	return EXIT_FAILURE;
      } 
	
    strlcpy(subject, _("Subject: "),            sizeof(subject));
    strlcat(subject, sh_unix_time (0),          sizeof(subject));
    strlcat(subject, " ",                       sizeof(subject));
    strlcat(subject, sh.localhost,              sizeof(subject));
    strlcat(subject, "\n\n",                    sizeof(subject));


    /* start writing
     */
    status = fwrite (subject, 1, strlen(subject), connfile); 
    status -= strlen(subject);

    /* ---------- Message ---------------------------------------- */

    strlcpy(subject, sh_unix_time (0),          sizeof(subject));
    strlcat(subject, " ",                       sizeof(subject));
    strlcat(subject, sh.localhost,              sizeof(subject));
    strlcat(subject, "\n",                      sizeof(subject));

    bufcompress = SH_ALLOC (SH_MAXBUF);
    bufcompress[0] = '\0';

    MLOCK(bufcompress , SH_MAXBUF);

    if (status == 0)
      {
	status = fwrite (_("<--- MESSAGE   ---->\n"), 
			 1, sizeof ("<--- MESSAGE   ---->\n")-1, 
			 connfile);
	status -= sizeof ("<--- MESSAGE   ---->\n")-1;
      }


    if (status == 0)
      {
	status = fwrite (subject, 1, strlen(subject), connfile);
	status -= strlen(subject);
      }
    sh_util_compress (bufcompress, subject, SH_MAXBUF-KEY_MAX-1);

    if (message != NULL)
      {
	if (status == 0) 
	  {
	    status = fwrite (message, 1, strlen(message), connfile);
	    status -= strlen(message);
	    if (status == 0) 
	      {
		status = fwrite ("\n", 1, 1, connfile);
		status -= 1;
	      }
	    sh_util_compress (bufcompress, message, SH_MAXBUF-KEY_MAX-1);
	  } else {
	    sh_mail_pushstack (message);  /* save for later */
	    ++failcount;
	    if (failcount > SH_MAX_FAIL) 
	      {
		sh_mail_emptystack ();
		failedMail = BAD;
	      }
	  }
      }

    while ( (NULL != (mailMsg = pop_list(fifo_mail))) && (status == 0) ) 
      {
	status = fwrite (mailMsg, 
			 1, strlen(mailMsg),
			 connfile);
	status -= strlen(mailMsg);
	if (status == 0) 
	  {
	    status = fwrite ("\n", 1, 1, connfile);
	    status -= 1;
	    sh_util_compress (bufcompress, 
			      mailMsg, 
			      SH_MAXBUF-KEY_MAX-1);
	    SH_FREE(mailMsg);
	    --sh.mailNum.alarm_last;
	  }
	else
	  {
	    push_tail_list (fifo_mail, mailMsg);
	  }
      }

    /* ------ signature block ---------------------------------- */

    if (status == 0)
      {
	status = fwrite (_("<--- SIGNATURE ---->\n"), 
			 1, sizeof ("<--- SIGNATURE ---->\n")-1,
			 connfile);
	status -= sizeof ("<--- SIGNATURE ---->\n")-1;
      }

    /* generate new signature key
     * - we generate it anyway; user can use mailcount to get the
     *   correct number of iterations --UNTESTED--
     */
    if (isfirst == 1) {
      sh_util_keyinit (skey->mailkey_old, KEY_MAX+1);
    }
    strlcpy(skey->mailkey_new, 
	    sh_tiger_hash (skey->mailkey_old, TIGER_DATA, KEY_MAX), 
	    KEY_MAX+1);
    
    /* First message is not signed. Next message is signed with the hash
     * of the signature key.
     */
    if (isfirst == 0) 
      {
	/* sign the message with the signature key
	 */
	strlcat (bufcompress, skey->mailkey_new, SH_MAXBUF);
	
	if (status == 0)
	  {
	    status = fwrite (sh_tiger_hash (bufcompress, 
					    TIGER_DATA,
					    strlen(bufcompress)),
			     1,
			     KEY_MAX,
			     connfile);
	    status -= KEY_MAX;
	  }
      } 
    else 
      {
	/* reveal first signature key
	 */
	strlcpy(skey->crypt, skey->mailkey_new, KEY_MAX+1);
	sh_util_encode(skey->crypt, bufcompress, 0);
	if (status == 0)
	  {
	    status = fwrite (skey->crypt,
			     1,
			     KEY_MAX, 
			     connfile);
	    status -= KEY_MAX;
	  }
	memset (skey->crypt, '\0', KEY_MAX);
	isfirst = 0;
      }
    
    /* X(n) -> X(n-1)
     */
    strlcpy (skey->mailkey_old, skey->mailkey_new, KEY_MAX+1);
 
    if (status == 0)
      {
	status = fwrite ("\n", 1, 1, connfile);
	status -= 1;
      }

    sprintf(subject, "%06d\n", mailcount);             /* known to fit  */
    if (status == 0)
      {
	status = fwrite (subject, 1, strlen(subject), connfile);
	status -= strlen(subject);
      }
    ++mailcount;

    if (status == 0)
      {
	status = fwrite (_("<--- END       ---->\n"), 
			 1,
			 sizeof ("<--- END       ---->\n")-1,
			 connfile);
	status -= sizeof ("<--- END       ---->\n")-1;
      }


    /* ------------- End -------------------------------------- */

    memset(bufcompress, '\0', SH_MAXBUF);

    MUNLOCK(bufcompress , SH_MAXBUF);

    SH_FREE(bufcompress);

    if (status == 0)
      sh_mail_end_conn (connfile);

    if (connfile != NULL) fclose (connfile);

    /* reset signal
     */
    sigaction (SIGPIPE, &old_act, NULL);

    if (status != EXIT_SUCCESS) {
      sh_error_handle (SH_ERR_FATAL, PLACE
		       NULL,
		       _("Mail error - broken connection."));
      return EXIT_FAILURE;
    } 

    return EXIT_SUCCESS;
}


/*
 *
 * SMTP CODE BELOW
 *
 *
 */

#include <ctype.h>
#include <netdb.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

/* missing on HP-UX 10.20 */
#ifndef IPPORT_SMTP
#define IPPORT_SMTP 25
#endif

static int sh_mail_wait(int code, FILE * m_socket);

static char * relay_host = NULL;

int sh_mail_set_relay (char * str_s)
{
  int i = 0;
  if (str_s == NULL)
    return -1;

  i = strlen(str_s);
  relay_host = malloc (i+1);
  if (relay_host != NULL)
    strlcpy(relay_host, str_s, i+1);
  else
    fprintf(stderr, _("ERROR:  sh_mail_set_relay: Out of memory"));
  return 0;
}


/*************************
 *
 * start connection
 * for details on SMTP, see RFC 821 
 */

static FILE * sh_mail_start_conn (char * address)
{

  char         this_address[256];
  char         m_machine[256];
  char         m_user[256];
  register int i, j, k;
  FILE       * connFile;

  int                  fd;
  struct in_addr       haddr;
  struct sockaddr_in   saddr;
  struct hostent     * hent = NULL; /* make compiler happy */
  char               * testit;
  int                  testflag = 0;


  /* -------   split adress ------------------  */
  if (strchr (address, '@') == NULL) {
    strlcpy(m_user,    address,     256);
    strlcpy(m_machine, _("localhost"), 256);
  } else {
    i = 0;
    while (i < 255 && address[i] != '@') {
      m_user[i] = address[i];
      ++i;
    }
    /* adress[i] = '@' */
    m_user[i] = '\0';
    j = i + 1; k = i; i = 0;
    while (i < 255 && address[i+j] != '\0') {
      m_machine[i] = address[i+j];
      ++i;
    }
    m_machine[i] = '\0';
    if (address[k] != '@' || address[k+i+1] != '\0') {
      return NULL;
    } 
  }


  if (relay_host != NULL) 
    strlcpy (m_machine, relay_host, sizeof(m_machine));
 
#ifdef MAIL_DEBUG
  fprintf(stderr,  _("user %s machine %s\n"), m_user, m_machine); 
#endif

  /* check whether it is in dotted number format
   */
  testit = &m_machine[0]; testflag = GOOD;
  while (*testit != '\0') 
    {
      if (!isdigit((int) *testit) && *testit != '.') testflag = BAD;
      ++testit;
    }

#ifdef MAIL_DEBUG
  if (testflag == BAD)
    fprintf(stderr,  _("no number-dot address: %s\n"), m_machine);
  else
    fprintf(stderr,  _("is a number-dot address: %s\n"), m_machine);
#endif


  /* host entry of other machine
   */
  if (testflag == GOOD)
    {
#ifdef HAVE_INET_ATON
      if (0 == inet_aton(m_machine, &haddr))
#else
      if ((unsigned long)-1  == (haddr.s_addr = inet_addr(m_machine))  )
#endif
	{
#ifdef MAIL_DEBUG
	  fprintf(stderr,  
		  _("inet_aton(%s) failed\n"), m_machine);
#endif
	  return NULL;
	}
    }
  else
    {
      if ( (hent = gethostbyname (m_machine)) == NULL) {
#ifdef MAIL_DEBUG
	fprintf(stderr,  _("gethostbyname(%s) failed\n"), m_machine); 
#endif
	return NULL;
      }
    }

  (void) memset (&saddr, 0, sizeof(saddr));
  if (testflag == GOOD)
    saddr.sin_addr   = haddr;
  else
    saddr.sin_addr   = *(struct in_addr *) hent->h_addr;
  saddr.sin_family = AF_INET;
  saddr.sin_port   = htons (IPPORT_SMTP);

  /* create a socket for communication
   */
  if ((fd = socket (AF_INET, SOCK_STREAM, IPPROTO_TCP)) == -1) {
#ifdef MAIL_DEBUG
  fprintf(stderr,  _("socket() failed\n"));
#endif
    return NULL;
  }
#ifdef MAIL_DEBUG
  fprintf(stderr,  _("created socket(): fd = %d\n"), fd);
#endif


  /* associate a FILE structure with it
   */
  if ((connFile = fdopen (fd, "r+")) == NULL) {
#ifdef MAIL_DEBUG
    fprintf(stderr,  _("fdopen() failed\n"));
#endif
    close(fd);
    return NULL;
  }

  /* connect our socket to another socket somewhere else
   */
  if (connect (fd, (struct sockaddr *) &saddr, sizeof(saddr))) {
#ifdef MAIL_DEBUG
    fprintf(stderr,  _("connect() failed\n"));
#endif
    fclose(connFile);
    connFile = NULL;
    return NULL;
  }



  /* say HELO to the other socket
   */
  if (!sh_mail_wait (220, connFile)) {
    fclose(connFile);
    return NULL;
  }
  fflush(connFile);
#ifdef MAIL_DEBUG
  fprintf(stderr,  _("HELO %s%c%c"), m_machine, 13, 10);
#endif
  fprintf(connFile, _("HELO %s%c%c"), m_machine, 13, 10);
  fflush(connFile);
  if (!sh_mail_wait(250, connFile)) {
#ifdef MAIL_DEBUG
    fprintf(stderr,  _("timeout\n"));
#endif
    fclose(connFile);  
    return NULL;
  }


  /* tell them who we are
   */
  strlcpy (this_address, DEFAULT_SENDER, 256);
  strlcat (this_address, "@", 256);
  strlcat (this_address, sh.localhost, 256);

#ifdef MAIL_DEBUG
  fprintf(stderr,  _("MAIL FROM:<%s>%c%c"), this_address, 13, 10);
#endif
  fflush(connFile);
  fprintf(connFile, _("MAIL FROM:<%s>%c%c"), this_address, 13, 10);
  fflush(connFile);
  if (!sh_mail_wait(250, connFile)) {
#ifdef MAIL_DEBUG
    fprintf(stderr,  _("timeout\n")); 
#endif
    fclose(connFile);
    return NULL;
  }

  /* tell them who to send mail to
   */
#ifdef MAIL_DEBUG
  fprintf(stderr,  _("RCPT TO:<%s>%c%c"), address, 13, 10); 
#endif
  fflush(connFile);
  fprintf(connFile, _("RCPT TO:<%s>%c%c"), address, 13, 10); 
  fflush(connFile);
  if (!sh_mail_wait(250, connFile)) {
#ifdef MAIL_DEBUG
    fprintf(stderr,  _("timeout\n")); 
#endif
    fclose(connFile);
    return NULL;
  }

  /* Send the message 
   */
#ifdef MAIL_DEBUG
  fprintf(stderr,  _("DATA%c%c"), 13, 10); 
#endif
  fflush(connFile);
  fprintf(connFile, _("DATA%c%c"), 13, 10);      
  fflush(connFile);
  if (!sh_mail_wait(354, connFile)) {
#ifdef MAIL_DEBUG
    fprintf(stderr,  _("timeout\n")); 
#endif
    fclose(connFile);
    return NULL;
  }

#ifdef MAIL_DEBUG
  fprintf(stderr,  
	  _("From: <%s>%c%c"\
	  "Date: %s%c%c"),
	  this_address, 13, 10,
	  sh_unix_time (0), 13, 10); 
#endif
  fflush(connFile);
  fprintf(connFile,
	  _("From: <%s>%c%c"\
	  "Date: %s%c%c"),
	  this_address, 13, 10,
	  sh_unix_time (0), 13, 10);

  return connFile;
}

/*************************
 *
 * end connection
 *
 */

static int sh_mail_end_conn (FILE * connFile)
{

  fflush(connFile);
  fprintf(connFile, _("%c%c.%c%c"), 13, 10, 13, 10);   
  fflush(connFile);
#ifdef MAIL_DEBUG
  fprintf(stderr,  _("message end written\n")); 
#endif

  if (sh_mail_wait(250, connFile))
    {  
      fflush(connFile);
      fprintf(connFile, _("QUIT%c%c"), 13, 10);
    }
#ifdef MAIL_DEBUG
  fprintf(stderr,  _("exit\n"));
#endif
    
  return (0);
}



/****************************
 *
 * Handle sever replies
 *
 *
 */



static jmp_buf wait_timeout;

static void sh_mail_alarmhandle (int signal)
{
  struct  sigaction  def_act;
  def_act.sa_handler = SIG_IGN;

  alarm(0);
  sigaction (SIGALRM, &def_act, NULL);
  longjmp(wait_timeout, 1);
}

static int sh_mail_wait(int code, FILE * m_socket)
{
  int rcode, g;

  enum { 
    WAIT_CODE_START, 
    WAIT_CODE, 
    WAIT_NL, 
    WAIT_NL_CONT 
  } state;

  struct  sigaction  def_act;
  struct  sigaction  new_act;
  def_act.sa_handler = SIG_IGN;
  new_act.sa_handler = sh_mail_alarmhandle;

  
  /* alarm was triggered
   */
  if (setjmp(wait_timeout) != 0)
    {
#ifdef MAIL_DEBUG
      fprintf(stderr,  _("mail_wait: timeout\n")); 
#endif
      return 0;
    }

  /* timeout after 5 minutes
   */
  alarm(0);
  sigaction (SIGALRM, &new_act, NULL);
  alarm(300);

#ifdef MAIL_DEBUG
  fprintf(stderr,  _("enter mail_wait\n")); 
#endif

  rcode = 0;
  state = WAIT_CODE_START;

  while (!feof(m_socket) && !ferror(m_socket)) {

    if ( (g=fgetc(m_socket)) == EOF)
      {
	alarm(0);
#ifdef MAIL_DEBUG
	fprintf(stderr,  _("mail_wait: EOF\n")); 
#endif
	sigaction (SIGALRM, &def_act, NULL); 
	return 0; 
      }

    switch(state) {

      /* wait for start of a numerical code
       */
    case WAIT_CODE_START:
      if (isspace(g))
	break;             /* Skip white space                    */
      if (!isdigit(g))
	return 0;          /* No leading number                   */
      rcode = g-'0';       /* convert to number                   */
      state = WAIT_CODE;
      break;
      
      /* wait for completion of numerical code
       */
    case WAIT_CODE:
      if (isdigit(g)) {
	rcode = rcode * 10 + (g-'0'); /* next digit               */
	break;
      }
      state = ((g == '-') ?  WAIT_NL_CONT :  WAIT_NL); 
      break;
      
      /* wait for newline, then return with status code
       */
    case WAIT_NL:
      if (g != '\n')
	break;
      alarm(0);                            /* Disable alarm       */
      sigaction (SIGALRM, &def_act, NULL); /* Reset handler       */
#ifdef MAIL_DEBUG
      fprintf(stderr,  _("mail_wait: OK got %d (%d) need %d (%d)\n"), 
	      rcode, (int)(rcode/100), code, (int)(code/100)); 
#endif
      return ((int)(rcode/100) == (int)(code/100)) ? 1 : 0;
      
      /* wait for continuation line
       */
    case WAIT_NL_CONT:
      if (g == '\n')
	state = WAIT_CODE_START;  /* There is a continuation line */
      break; 
      
    default:
      alarm(0);                            /* Disable alarm       */
      sigaction (SIGALRM, &def_act, NULL); /* Reset handler       */
#ifdef MAIL_DEBUG
      fprintf(stderr,  _("mail_wait: bad\n")); 
#endif
      return 0;
      
    }
  }

  alarm(0);                            /* Disable alarm       */
  sigaction (SIGALRM, &def_act, NULL); /* Reset handler       */
#ifdef MAIL_DEBUG
  fprintf(stderr,  _("mail_wait: failed\n")); 
#endif
  return 0;                      /* Failed, EOF or error on socket */
}

  


