/* SAMHAIN file system integrity testing                                   */
/* Copyright (C) 1999 Rainer Wichmann                                      */
/*                                                                         */
/*  This program is free software; you can redistribute it                 */
/*  and/or modify                                                          */
/*  it under the terms of the GNU General Public License as                */
/*  published by                                                           */
/*  the Free Software Foundation; either version 2 of the License, or      */
/*  (at your option) any later version.                                    */
/*                                                                         */
/*  This program is distributed in the hope that it will be useful,        */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/*  GNU General Public License for more details.                           */
/*                                                                         */
/*  You should have received a copy of the GNU General Public License      */
/*  along with this program; if not, write to the Free Software            */
/*  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.              */



#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "config_xor.h"

#ifdef HAVE_MEMORY_H
#include <memory.h>
#endif

#include "samhain.h"
#include "sh_error.h"
#include "sh_utils.h"
#include "sh_mem.h"


#ifdef MEM_DEBUG

#define CHECKBYTE 0x7F

typedef struct mem_struct {
  struct mem_struct *next;        /* link to next struct */
  char * address;                 /* address assigned    */
  unsigned long size;             /* size allocated      */
  char file[20];                  /* Allocation file name */
  int line;                       /* Allocation line number */
} memlist_t;

typedef struct open_struct {
  struct open_struct *next;
  char   file[50];
  int    fd;
} openfiles_t;

memlist_t   * memlist       = NULL;
openfiles_t * openfileslist = NULL;

int Free_Count = 0, Alloc_Count = 0;
unsigned long Mem_Current = 0, Mem_Max = 0;

void sh_mem_openf (char * file, int fd)
{
  openfiles_t * this;

  if ( (this = malloc (sizeof(openfiles_t))) == NULL) {
    sh_error_handle (SH_ERR_FAIL, PLACE
		     NULL, _("Out of memory") );
    return;
  }

  /* make list entry */

  this->fd      = fd;
  strlcpy(this->file, file, 50);

  this->next = openfileslist;
  openfileslist = this;

  return;
}

void sh_mem_closef (int fd)
{
  openfiles_t * this   = openfileslist;
  openfiles_t * before = openfileslist;

  /* find record */
  while (this != NULL) {
    if (this->fd == fd) break;
    before = this;
    this   = this->next;
  }

  if (this == NULL) {
    sh_error_handle (SH_ERR_ERR, PLACE
		     _("Close() on unrecorded file"),
		     _("Close() on unrecorded file - fd %d"),
		     fd);
  } else {
    if (this == openfileslist) 
      openfileslist = this->next;
    else 
      before->next  = this->next;
    free (this);
  }
  return;
}

void sh_mem_stat ()
{
  memlist_t   * this;
  openfiles_t * file;

  if (Alloc_Count == Free_Count  &&  sh.openfiles == 0) {
    sh_error_handle (SH_ERR_STAMP, PLACE
		     NULL, 
		     _("Memory used:  Max=%ld, Now=%ld; Open Files: %d"),
		     Mem_Max, Mem_Current, sh.openfiles);
    return;
  }
    
  sh_error_handle (SH_ERR_INFO, PLACE
		   NULL, _("Allocated Blocks: %d, Free'd Blocks: %d"),
		   Alloc_Count, Free_Count);
  sh_error_handle (SH_ERR_INFO, PLACE
		   NULL, _("Allocated (bytes): Maximum=%ld, Current=%ld"),
		   Mem_Max, Mem_Current);
  sh_error_handle (SH_ERR_INFO, PLACE
		   NULL, _("Open files: %d"),
		   sh.openfiles);

  this = memlist;
  file = openfileslist;

  while (this != NULL) {
    sh_error_handle (SH_ERR_INFO, PLACE
		     NULL, 
		     _("Not deallocated: size %14ld, file %19s, line %d"),
		     this->size, this->file, this->line);
    this = this->next;
  }

  while (file != NULL) {
    sh_error_handle (SH_ERR_INFO, PLACE 
		     NULL, 
		     _("Open file: fd %d, file %19s"),
		     file->fd, file->file);
    file = file->next;
  }

  return;
}

void sh_mem_check ()
{
  memlist_t * this;

  sh_error_handle (SH_ERR_STAMP, PLACE
		   NULL, 
		   _("Memory used:  Max=%ld, Now=%ld; Open Files: %d"),
		   Mem_Max, Mem_Current, sh.openfiles);

  this = memlist;

  while (this != NULL) {
    if ( this->address[this->size] != CHECKBYTE ) {
      sh_error_handle (SH_ERR_ERR, PLACE
		       NULL,
		       _("Memory overrun on block allocated "\
			 "- file %s, line %d"),
		       this->file, 
		       this->line);
      }
    this = this->next;
  }

  return;
}

void * sh_mem_malloc (unsigned long size, char * file, int line)
{
  void      * theAddress;
  memlist_t * this;
  static char errormsg[42];

  if ( (theAddress = malloc(size+1)) == NULL ) {
    strcpy(errormsg, _("Out of memory - file "));       /* Known to fit  */
    strlcat(errormsg, file, 42);

    sh_error_handle (SH_ERR_FAIL, PLACE
		     _("Out of memory"),
		     errormsg);
    /* use _exit() rather than exit() - we malloc() in the atexit() functions 
     */
    _exit (EXIT_FAILURE);
  }

  /* set check bytes */
  memset(theAddress, CHECKBYTE, size + 1);
  memset(theAddress, 0, 1);

  ++Alloc_Count;

  Mem_Current += size;
  Mem_Max = ( (Mem_Current > Mem_Max) ? Mem_Current : Mem_Max);

  if ( (this = (memlist_t *) malloc (sizeof(memlist_t))) == NULL) {
    strcpy(errormsg, _("Out of memory - file "));       /* Known to fit  */
    strlcat(errormsg, file, 42);

    sh_error_handle (SH_ERR_FAIL, PLACE
		     _("Out of memory"),
		     errormsg);
    return theAddress;
  }

  /* make list entry */

  this->address = theAddress;
  this->size    = size;
  this->line    = line;
  strlcpy(this->file, file, 20);

  this->next = memlist;
  memlist = this;

  return theAddress;
}


void sh_mem_free (void * a, char * file, int line)
{
  memlist_t * this   = memlist;
  memlist_t * before = memlist;

  unsigned long size = 0;

  if ( a == NULL ) {
    sh_error_handle (SH_ERR_ERR, PLACE
		     _("Dereferenced NULL pointer"),
		     _("Dereferenced NULL pointer - file %s, line %d"),
		     file, 
		     line);
    return;
  }
    
  /* find record */
  while (this != NULL) {
    if (this->address == a) 
      break;
    before = this;
    this   = this->next;
  }

  if (this == NULL) {
    sh_error_handle (SH_ERR_ERR, PLACE
		     _("Free() on unrecorded block"),
		     _("Free() on unrecorded block - file %s, line %d"),
		     file, 
		     line);
  } else {
    if ( this->address[this->size]  != CHECKBYTE )
      sh_error_handle (SH_ERR_ERR, PLACE
		       _("Memory overrun"),
		       _("Memory overrun on block (allocated %s, %d) - file %s, line %d"),
		       file, line,
		       this->file, 
		       this->line);

    size = this->size;

    if (this == memlist) 
      memlist = this->next;
    else 
      before->next = this->next;
    free (this);
  }

  free (a);
  ++Free_Count;
  Mem_Current -= size;
  return;
}
#else
void sh_mem_free (void * a)
{
  if ( a == NULL ) 
    sh_error_handle (SH_ERR_ERR, PLACE
		     NULL,
		     _("Dereferenced NULL pointer"));
  else
    free (a);
  return;
}
void * sh_mem_malloc (unsigned long size)
{
  void * theAddress;

  if ( (theAddress = malloc(size)) == NULL ) {
    sh_error_handle (SH_ERR_FAIL, PLACE
		     NULL,
		     _("Out of memory"));
    /* use _exit() rather than exit() - we malloc() in the atexit() functions 
     */
    _exit (EXIT_FAILURE);
  }

  return theAddress;
}
#endif
