/* SAMHAIN file system integrity testing                                   */
/* Copyright (C) 1999 Rainer Wichmann                                      */
/*                                                                         */
/*  This program is free software; you can redistribute it                 */
/*  and/or modify                                                          */
/*  it under the terms of the GNU General Public License as                */
/*  published by                                                           */
/*  the Free Software Foundation; either version 2 of the License, or      */
/*  (at your option) any later version.                                    */
/*                                                                         */
/*  This program is distributed in the hope that it will be useful,        */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/*  GNU General Public License for more details.                           */
/*                                                                         */
/*  You should have received a copy of the GNU General Public License      */
/*  along with this program; if not, write to the Free Software            */
/*  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.              */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>

#include "config_xor.h"

#include "samhain.h"
#include "sh_error.h"
#include "sh_unix.h"
#include "sh_utils.h"
#include "sh_files.h"
#include "sh_mail.h"
#include "sh_forward.h"
#include "sh_modules.h"


typedef enum {
  SH_SECTION_NONE,
  SH_SECTION_LOG,
  SH_SECTION_MISC,
  SH_SECTION_READONLY,
  SH_SECTION_LOGFILES,
  SH_SECTION_LOGGROW,
  SH_SECTION_NOIGNORE,
  SH_SECTION_ALLIGNORE,
#if defined (SH_WITH_CLIENT) || defined (SH_STANDALONE) 
  SH_SECTION_OTHER,
#endif
#ifdef SH_WITH_SERVER
  SH_SECTION_CLIENTS,
#endif
  SH_SECTION_THRESHOLD
} ShSectionType;
    
static void sh_readconfig_line (char * line);

static ShSectionType read_mode = SH_SECTION_NONE;

/* read the configuration file
 */
int sh_readconf_read ()
{
#if defined (SH_WITH_CLIENT) || defined (SH_STANDALONE) 
  int    modnum;
#endif

  int    fd;
  char * tmp;
  char   line[512+2];
  int    some_other_host = 0;
  int    length = strlen(sh.localhost);

  /* open config file, exit on failure
   */
  if (BAD == tf_trust_check (DEFAULT_CONFIGFILE))
    exit (EXIT_FAILURE);

  if ( (fd = sh_unix_fileopen (DEFAULT_CONFIGFILE, 'r', 
			       SH_ISDATA, SH_ERR_FATAL, YESPRIV)) == -1) {
    sh_error_handle (SH_ERR_FATAL, PLACE
		     NULL,
		     _("Cannot open configuration file %s"),
		     DEFAULT_CONFIGFILE);
    exit (EXIT_FAILURE);
  }

#ifndef SH_STEALTH
  while ( sh_unix_getline (fd, line, 512) > 0) {
#else
  while ( sh_unix_getline_stealth (fd, line, 512) > 0) {
#endif

    /* skip header etc. */
    if (line[0] == '#'  ||  
	line[0] == ' '  || 
	line[0] == '\n' || 
	line[0] == '\0' || 
	line[0] == ';'  || 
	line[0] == '/' ) continue; 
  
    /* clip off trailing white space                 */
    tmp = line + strlen( line ); --tmp;
    while( isspace((int) *tmp ) && tmp >= line ) *tmp-- = '\0';

    /* -------  starts a section  ------------  */
    
    if (line[0] == '[' && some_other_host == 0) 
      { 

	if      (strncmp (line,  _("[ReadOnly]"), 
			  sizeof("[ReadOnly]")-1) == 0) 
	  read_mode = SH_SECTION_READONLY;
	else if (strncmp (line,  _("[LogFiles]"), 
			  sizeof("[LogFiles]")-1) == 0) 
	  read_mode = SH_SECTION_LOGFILES;
	else if (strncmp (line,  _("[GrowingLogFiles]"), 
			  sizeof("[GrowingLogFiles]")-1) == 0) 
	  read_mode = SH_SECTION_LOGGROW;
	else if (strncmp (line,  _("[IgnoreAll]"), 
			  sizeof("[IgnoreAll]")-1) == 0) 
	  read_mode = SH_SECTION_ALLIGNORE;
	else if (strncmp (line,  _("[IgnoreNone]"), 
			  sizeof("[IgnoreNone]")-1) == 0) 
	  read_mode = SH_SECTION_NOIGNORE;
	else if (strncmp (line,  _("[Log]"), 
			  sizeof("[Log]")-1) == 0) 
	  read_mode = SH_SECTION_LOG;
	else if (strncmp (line,  _("[Misc]"), 
			  sizeof("[Misc]")-1) == 0) 
	  read_mode = SH_SECTION_MISC;
	else if (strncmp (line,  _("[EOF]"), 
			  5) == 0)
	  {
	    sh_unix_fileclose (fd);
	    return (EXIT_SUCCESS);
	  }
	else if (strncmp (line,  _("[EventSeverity]"), 
			  sizeof("[EventSeverity]")-1) == 0) 
	  read_mode = SH_SECTION_THRESHOLD;

#ifdef SH_WITH_SERVER
	else if (strncmp (line,  _("[Clients]"), 
			  sizeof("[Clients]")-1) == 0) 
	  read_mode = SH_SECTION_CLIENTS;
#endif

#if defined (SH_WITH_CLIENT) || defined (SH_STANDALONE) 
	else
	  for (modnum = 0; modList[modnum].name != NULL; ++modnum) 
	    {
	      if (0 == strncmp (line,  modList[modnum].conf_section,
				strlen(modList[modnum].conf_section)) )
		read_mode = SH_SECTION_OTHER;
	    }
#endif
      } 

    /* ---  an @host directive -------------- */

    else if (line[0] == '@') 
      {
	if (strncmp (&line[1], "end", 3) == 0)
	  {
	    some_other_host = 0;   /* end  of an @host directive */
	  }
	else if  (strncmp (&line[1],  sh.localhost, length) != 0)
	  {
	    some_other_host = 1;   /* start of an @host directive */
	  }
      } 

    /* ------  no new section -------------- */

    else if (some_other_host == 0 && read_mode != SH_SECTION_NONE) 
      { 
	sh_readconfig_line (line);
      }
  }

  sh_unix_fileclose (fd);
  return (EXIT_SUCCESS);
}

typedef enum {
  SET_MAILTIME,
  SET_FILETIME 
} ShTimerItem;

    
int sh_readconf_setTime (char * str, ShTimerItem what)
{
  unsigned long i = atoi (str);

  if (i >= 0 && i < LONG_MAX) {
    if      (what == SET_MAILTIME) sh.mailTime.alarm_interval = i;
    else if (what == SET_FILETIME) sh.fileCheck.alarm_interval  = i;
    return GOOD;
  } else {
    sh_error_handle (SH_ERR_ERR, PLACE
		     NULL,
		     _("Invalid argument for maximum time interval (%s)"),
		     str);
    return BAD;
  }
}

static void sh_readconfig_line (char * line)
{    
  char * c;
  char * tmp;
  int    i;
#if defined (SH_WITH_CLIENT) || defined (SH_STANDALONE) 
  int    modnum, modkey;
#endif
  int    key = 0;

  static char  *ident[] = {
    N_("SeverityReadOnly"),
    N_("SeverityLogFiles"),
    N_("SeverityGrowingLogs"),
    N_("SeverityIgnoreNone"),
    N_("SeverityIgnoreAll"),
    N_("SeverityDirs"),
    N_("SeverityFiles"),
    N_("SeverityNames"),
    NULL
  };

  static int      identnum[] = { 
  SH_ERR_T_RO,    
  SH_ERR_T_LOGS,  
  SH_ERR_T_GLOG,  
  SH_ERR_T_NOIG,  
  SH_ERR_T_ALLIG, 
  SH_ERR_T_DIR,   
  SH_ERR_T_FILE, 
  SH_ERR_T_NAME        
  };
    

  /* interpret line */
  c = strchr(line, '=');
  if (c == NULL) return; /* no 'key=value' pair  */
  ++c;
  if ((*c) == '\0' || (*c) == ' ') return; /* no value */

  /* clip off trailing white space                 */
  tmp = line + strlen( line ); --tmp;
  while( isspace( (int) *tmp ) && tmp >= line ) *tmp-- = '\0';

  if (sh_unix_isnotsuid())
    sh_error_handle (SH_ERR_ALL, PLACE
		     NULL, _("ConfigFile: %s"), line);

  /* file or directory                             */
  if      (strncmp (line, _("dir"),  sizeof("dir")-1)  == 0) 
    key = 2;
  else if (strncmp (line, _("file"), sizeof("file")-1) == 0) 
    key = 1;
  else
    key = 0;

#if defined (SH_WITH_CLIENT) || defined (SH_STANDALONE) 
    
  if      (read_mode == SH_SECTION_READONLY) {
    if      (key == 1) sh_files_pushfile_ro (c);
    else if (key == 2) sh_files_pushdir_ro  (c);
  }
  else if (read_mode == SH_SECTION_LOGFILES) {
    if      (key == 1) sh_files_pushfile_log (c);
    else if (key == 2) sh_files_pushdir_log  (c);
  }
  else if (read_mode == SH_SECTION_LOGGROW) {
    if      (key == 1) sh_files_pushfile_glog (c);
    else if (key == 2) sh_files_pushdir_glog  (c);
  }
  else if (read_mode == SH_SECTION_NOIGNORE) {
    if      (key == 1) sh_files_pushfile_noig (c);
    else if (key == 2) sh_files_pushdir_noig  (c);
  }
  else if (read_mode == SH_SECTION_ALLIGNORE) {
    if      (key == 1) sh_files_pushfile_allig (c);
    else if (key == 2) sh_files_pushdir_allig  (c);
  }

  else if (read_mode == SH_SECTION_MISC) {
#else
  if (read_mode == SH_SECTION_MISC) {
#endif
    if (strncmp (line,  _("SetFilecheckTime"), 
		 sizeof  ("SetFilecheckTime")-1) == 0) {
      sh_readconf_setTime (c, SET_FILETIME);
    }
    else if (strncmp (line,  _("SetMailTime"), 
		      sizeof  ("SetMailTime")-1) == 0) {
      sh_readconf_setTime (c, SET_MAILTIME);
    }
    else if (strncmp (line,  _("SetMailNum"), 
		      sizeof  ("SetMailNum")-1) == 0) {
      sh_mail_setNum (c);
    }
    else if (strncmp (line,  _("SetMailAddress"), 
		      sizeof  ("SetMailAddress")-1) == 0) {
      sh_mail_setaddress (c);
    }
    else if (strncmp (line,  _("SetMailRelay"), 
		      sizeof  ("SetMailRelay")-1) == 0) {
      sh_mail_set_relay (c);
    }
    else if (strncmp (line,  _("TrustedUser"), 
		      sizeof  ("TrustedUser")-1) == 0) {
      tf_add_trusted_user (c);
    }
    else if (strncmp (line,  _("SetTimeServer"), 
		      sizeof  ("SetTimeServer")-1) == 0) {
      sh_unix_settimeserver (c);
    }
#ifdef SH_WITH_CLIENT
    else if (strncmp (line,  _("SetLogServer"), 
		      sizeof("SetLogServer")-1) == 0) {
      sh_forward_setlogserver (c);
    }
#endif
    else if (strncmp (line,  _("SetLoopTime"), 
		      sizeof("SetLoopTime")-1) == 0) {
      sh_util_setlooptime (c);
    }
#if defined (SH_WITH_CLIENT) || defined (SH_STANDALONE) 
    else if (strncmp (line,  _("SetRecursionLevel"), 
		      sizeof("SetRecursionLevel")-1) == 0) {
      sh_files_setrecursion (c);
    }
    else if (strncmp (line,  _("ChecksumTest"), 
		      sizeof("ChecksumTest")-1) == 0) {
      sh_util_setchecksum (c);
    }
#endif
  }

#if defined (SH_WITH_CLIENT) || defined (SH_STANDALONE) 
  else if      (read_mode == SH_SECTION_OTHER) {
    for (modnum = 0; modList[modnum].name != NULL; ++modnum) 
      {
	for (modkey = 0; modList[modnum].conf_table[modkey].the_opt != NULL; 
	     ++modkey) 
	  {
	    if (strncmp (line,  modList[modnum].conf_table[modkey].the_opt,
		     strlen(modList[modnum].conf_table[modkey].the_opt)) == 0)
	      modList[modnum].conf_table[modkey].func(c);
	  }
      }
  }
#endif


#ifdef SH_WITH_SERVER
  else if (read_mode == SH_SECTION_CLIENTS) {
    if      (strncmp (line,  _("Client"), 
		      sizeof("Client")-1) == 0) 
      sh_forward_register_client (c);
  }
#endif

  else if (read_mode == SH_SECTION_LOG) {
    if      (strncmp (line,  _("DebugLevel"), 
		      sizeof("DebugLevel")-1) == 0) 
      sh_error_setdebug (c);
    else if (strncmp (line,  _("MailSeverity"), 
		      sizeof("MailSeverity")-1) == 0)
      sh_error_setseverity (c);
    else if (strncmp (line,  _("PrintSeverity"), 
		      sizeof("PrintSeverity")-1) == 0)
      sh_error_setprint (c);
#ifdef SH_WITH_CLIENT
    else if (strncmp (line,  _("ExportSeverity"), 
		      sizeof("ExportSeverity")-1) == 0)
      sh_error_setexport (c);
#endif
    else if (strncmp (line,  _("LogSeverity"), 
		      sizeof("LogSeverity")-1) == 0)
      sh_error_setlog (c);
    else if (strncmp (line,  _("SyslogSeverity"), 
		      sizeof("SyslogSeverity")-1) == 0)
      sh_error_set_syslog (c);
  }
  
  else if (read_mode == SH_SECTION_THRESHOLD) {
    i = 0;
    while (ident[i] != NULL) {
      if (strncmp (line, _(ident[i]), strlen(ident[i])-1) == 0)
	sh_error_set_iv (identnum[i], c);
      ++i;
    }
  }
     
  return;
}
  
    
