/* SAMHAIN file system integrity testing                                   */
/* Copyright (C) 1999 Rainer Wichmann                                      */
/*                                                                         */
/*  This program is free software; you can redistribute it                 */
/*  and/or modify                                                          */
/*  it under the terms of the GNU General Public License as                */
/*  published by                                                           */
/*  the Free Software Foundation; either version 2 of the License, or      */
/*  (at your option) any later version.                                    */
/*                                                                         */
/*  This program is distributed in the hope that it will be useful,        */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/*  GNU General Public License for more details.                           */
/*                                                                         */
/*  You should have received a copy of the GNU General Public License      */
/*  along with this program; if not, write to the Free Software            */
/*  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.              */

#include <stdio.h>
#include <stdlib.h>

#include "config_xor.h"

#include "samhain.h"

#ifdef USE_SRP_PROTOCOL

#if (defined (SH_WITH_CLIENT) || defined (SH_WITH_SERVER))

#include "sh_tiger.h"
#include "sh_mem.h"
#include "sh_utils.h"
#include "bignum.h"

extern char TcpFlag[9];

typedef struct sh_srp_struc {
  char   x[KEY_MAX+1];
  bignum a;
  bignum p;
  bignum g;
} sh_srp_t;

static sh_srp_t sh_srp;

void sh_srp_x (char * salt, char * password)
{

  char           *combi;
  long            len;
  register int i;
  unsigned char * dez = NULL;

  if (password == NULL)
    dez = (unsigned char *) &TcpFlag[0];
  else if (strlen(password) >= 8)
    dez = (unsigned char *) password;
  else
    {
      fprintf(stderr, _("Password has less than 8 chars !\n"));
      _exit(EXIT_FAILURE);
    }

  for (i = 0; i < 8; ++i)
    {
      skey->vernam[i] = (*dez != '\0' ?  *dez : '\001'); ++dez;
      skey->vernam[i+1] = '\0';
    }
  skey->vernam[8] = '\0';
  strlcpy (skey->vernam,
	   sh_tiger_hash(skey->vernam, TIGER_DATA, 8), KEY_MAX);
  skey->vernam[16] = skey->vernam[48]; 
  skey->vernam[33] = skey->vernam[49]; 
  skey->vernam[48] = '\0';

  len = strlen(salt) + strlen(skey->vernam) + 1;

  /* H(s,P)
   */
  combi = SH_ALLOC(len);
  strlcpy (combi, salt, len);
  strlcat (combi, skey->vernam, len);
  strlcpy (sh_srp.x, 
	   sh_tiger_hash(combi, TIGER_DATA, strlen(combi)),
	   KEY_MAX+1);
  SH_FREE (combi);
  sh_srp.x[16] = sh_srp.x[48]; 
  sh_srp.x[33] = sh_srp.x[49]; 
  sh_srp.x[48] = '\0';
  return;
}

char * sh_srp_M (char * x1, char * x2, char * x3)
{
    char           *combi;
    long            len;
    static char     hash[KEY_MAX+1];

    len = strlen(x1) + strlen(x2) + strlen(x3) + 1;

    /* H(x1,x2,x3)
     */
    combi = SH_ALLOC(len);
    strlcpy (combi, x1, len);
    strlcat (combi, x2, len);
    strlcat (combi, x3, len);
    strlcpy (hash, 
	     sh_tiger_hash(combi, TIGER_DATA, strlen(combi)),
	     KEY_MAX+1);
    SH_FREE (combi);

    return hash;
}

int sh_srp_init()
{
  bigerr_t res;
  char     modulus[80*4];

  modulus[0] = '\0';

  res = big_init_pkg();
  
  if (res == BIG_OK)
    {
      big_create(&sh_srp.p);
      big_create(&sh_srp.g);
      big_create(&sh_srp.a);
      strlcpy(modulus, SRP_MODULUS_1024_1, sizeof(modulus));
      strlcat(modulus, SRP_MODULUS_1024_2, sizeof(modulus));
      strlcat(modulus, SRP_MODULUS_1024_3, sizeof(modulus));
      strlcat(modulus, SRP_MODULUS_1024_4, sizeof(modulus));
      big_set_string (modulus,                  16, &sh_srp.p);
      big_set_string (SRP_GENERATOR_1024,       16, &sh_srp.g);
      return GOOD;
    }
  return BAD;
}

void sh_srp_exit()
{
  big_destroy(&sh_srp.g);	     
  big_destroy(&sh_srp.p);
  big_destroy(&sh_srp.a);

  big_release_pkg();
}

void sh_srp_make_a ()
{
  char randy[5];
  char hash[KEY_MAX+1];

  sh_util_keyinit (randy, 5);
  strlcpy (hash, sh_tiger_hash(randy, TIGER_DATA, 8), KEY_MAX);

  hash[16] = hash[48]; 
  hash[33] = hash[49]; 
  hash[48] = '\0';
  big_set_string (hash,       16, &sh_srp.a);
    
  return;
}

int sh_srp_check_zero (char * AB_str)
{
  bignum   AB, q, r;
  bigerr_t res;
  int      val;

  big_create(&AB);
  big_create(&q);
  big_create(&r);

  big_set_string (AB_str,       16, &AB);
  res = big_trunc(&AB, &sh_srp.p, &q, &r); /* is last one the remainder ? */
  
  if ( big_zerop(&AB) ) val = BAD;
  else                  val = GOOD;

  big_destroy(&AB);	
  big_destroy(&q);	
  big_destroy(&r);	
  
  return val;
}

#ifdef SH_WITH_CLIENT
  
char * sh_srp_A ()
{
  bignum   A;
  char    *str;
  char    *combi;
  long     len;
  bigerr_t res;


  big_create(&A);
  
  res = big_exptmod (&sh_srp.g, &sh_srp.a, &sh_srp.p, &A);
  
  str = big_string (&A, 16);
  
  len = strlen(str) + 1;
  combi = SH_ALLOC(len);
  strlcpy (combi, str, len);
  
  big_destroy(&A);	     
  return combi;
}

/* #ifdef SH_WITH_CLIENT */
#endif  
  
#ifdef SH_WITH_SERVER

char * sh_srp_B (char * verifier)
{
  bignum   B, v, t, dummy;
  char    *str;
  char    *combi;
  long     len;
  bigerr_t res;

  big_create(&dummy);
  big_create(&t);
  big_create(&v);
  big_create(&B);
  
  res = big_exptmod (&sh_srp.g, &sh_srp.a, &sh_srp.p, &t);
  
  big_set_string (verifier,       16, &v);

  res = big_add (&t, &v, &dummy);

  if ( big_greaterp(&dummy, &sh_srp.p) ) big_sub(&dummy, &sh_srp.p, &B);
  else big_set_big(&dummy, &B);

  str = big_string (&B, 16);
  
  len = strlen(str) + 1;
  combi = SH_ALLOC(len);
  strlcpy (combi, str, len);
  
  big_destroy(&B);	
  big_destroy(&v);	
  big_destroy(&t);	
  big_destroy(&dummy);	
  
  return combi;
}
/* #ifdef SH_WITH_SERVER */
#endif  
  
  
#ifdef SH_WITH_CLIENT
  
char * sh_srp_S_c (char * u_str, char * B_str)
{
  bignum   u, B, x, t, base, z1, z2;
  char    *str;
  char    *combi;
  long     len;
  bigerr_t res;

  big_create(&z2);
  big_create(&z1);
  big_create(&base);
  big_create(&t);
  big_create(&x);
  big_create(&B);
  big_create(&u);
  
  big_set_string (B_str,          16, &B);
  big_set_string (sh_srp.x,       16, &x);
  big_set_string (u_str,          16, &u);
  
  /* the base  (B - g^x)
   */
  res = big_exptmod (&sh_srp.g, &x, &sh_srp.p, &t);
  if ( big_greaterp(&B, &t) ) 
    {
      res = big_sub(&B, &t, &base);
    }
  else 
    {
      res = big_add(&B, &sh_srp.p, &z2);
      res = big_sub(&z2, &t, &base);
    }

  /* the exponent (a + ux)
   */
  res = big_mul (&u, &x, &t);
  res = big_trunc(&t, &sh_srp.p, &z1, &z2); /* is last one the remainder ? */
  res = big_add(&sh_srp.a, &z2, &z1);
  if ( big_greaterp(&z1, &sh_srp.p) ) 
    {
      res = big_sub(&z1, &sh_srp.p, &z2);
    }
  else 
    {
      big_set_big(&z1, &z2);
    }

  res = big_exptmod (&base, &z2, &sh_srp.p, &t);

  str = big_string (&t, 16);
  
  len = strlen(str) + 1;
  combi = SH_ALLOC(len);
  strlcpy (combi, str, len);
  
  big_destroy(&z1);	     
  big_destroy(&z2);	     
  big_destroy(&base);	     
  big_destroy(&t);	
  big_destroy(&x);	
  big_destroy(&B);	
  big_destroy(&u);	
  
  return combi;
}
  
/* #ifdef SH_WITH_CLIENT */
#endif  
  
#ifdef SH_WITH_SERVER

  
char * sh_srp_S_s (char * u_str, char * A_str, char * v_str)
{
  bignum   u, A, v, t, base, z1, z2;
  char    *str;
  char    *combi;
  long     len;
  bigerr_t res;

  big_create(&z2);
  big_create(&z1);
  big_create(&base);
  big_create(&t);
  big_create(&v);
  big_create(&A);
  big_create(&u);
  
  big_set_string (A_str,          16, &A);
  big_set_string (v_str,          16, &v);
  big_set_string (u_str,          16, &u);
  
  /* the base  (Av^u)
   */
  res = big_exptmod (&v, &u, &sh_srp.p, &t);
  res = big_mul (&A, &t, &z1);
  res = big_trunc(&z1, &sh_srp.p, &z2, &base); /* is last the remainder ? */

  res = big_exptmod (&base, &sh_srp.a, &sh_srp.p, &t);

  str = big_string (&t, 16);
  
  len = strlen(str) + 1;
  combi = SH_ALLOC(len);
  strlcpy (combi, str, len);
  
  big_destroy(&z1);	     
  big_destroy(&z2);	     
  big_destroy(&base);	     
  big_destroy(&t);	
  big_destroy(&v);	
  big_destroy(&A);	
  big_destroy(&u);	
  
  return combi;
}

/* #ifdef SH_WITH_SERVER */
#endif  

char * sh_srp_verifier ()
{
  bignum   x, v;
  char    *combi;
  char    *str;
  long     len;
  bigerr_t res;
  
  big_create(&x);
  big_create(&v);
  
  big_set_string (sh_srp.x,               16, &x);
  
  res = big_exptmod (&sh_srp.g, &x, &sh_srp.p, &v);
  
  str = big_string (&v, 16);
  
  len = strlen(str) + 1;
  combi = SH_ALLOC(len);
  strlcpy (combi, str, len);
  
  big_destroy(&x);	     
  big_destroy(&v);	     
  
  return combi;
}
  

/* #if (defined (SH_WITH_CLIENT) || defined (SH_WITH_SERVER)) */

#endif

/* #ifdef USE_SRP_PROTOCOL */

#endif




