/* SAMHAIN file system integrity testing                                   */
/* Copyright (C) 1999 Rainer Wichmann                                      */
/*                                                                         */
/*  This program is free software; you can redistribute it                 */
/*  and/or modify                                                          */
/*  it under the terms of the GNU General Public License as                */
/*  published by                                                           */
/*  the Free Software Foundation; either version 2 of the License, or      */
/*  (at your option) any later version.                                    */
/*                                                                         */
/*  This program is distributed in the hope that it will be useful,        */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/*  GNU General Public License for more details.                           */
/*                                                                         */
/*  You should have received a copy of the GNU General Public License      */
/*  along with this program; if not, write to the Free Software            */
/*  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.              */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "config_xor.h"

#ifdef HAVE_MEMORY_H
#include <memory.h>
#endif

#ifdef  HAVE_UNISTD_H
#include <errno.h>
#include <signal.h>
#include <pwd.h>
#include <grp.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/resource.h>
#include <fcntl.h>
#include <unistd.h>

/*********************
#ifdef HAVE_SYS_VFS_H
#include <sys/vfs.h>
#endif
**********************/
#endif

#if TIME_WITH_SYS_TIME
#include <sys/time.h>
#include <time.h>
#else
#if HAVE_SYS_TIME_H
#include <sys/time.h>
#else
#include <time.h>
#endif
#endif

#ifdef HAVE_SYS_SELECT_H
#include <sys/select.h>
#endif

#ifndef FD_SET
#define NFDBITS         32
#define FD_SETSIZE      32
#define FD_SET(n, p)    ((p)->fds_bits[(n)/NFDBITS] |= (1 << ((n) % NFDBITS)))
#define FD_CLR(n, p)    ((p)->fds_bits[(n)/NFDBITS] &= ~(1 << ((n) % NFDBITS)))
#define FD_ISSET(n, p)  ((p)->fds_bits[(n)/NFDBITS] & (1 << ((n) % NFDBITS)))
#define FD_ZERO(p)      memset((char *)(p), '\0', sizeof(*(p)))
#endif /* !FD_SET */


#if defined(HAVE_MLOCK) && !defined(HAVE_BROKEN_MLOCK)
#include <sys/mman.h>
#endif

#include "samhain.h"
#include "sh_error.h"
#include "sh_unix.h"
#include "sh_utils.h"
#include "sh_mem.h"

#include "trustfile.h"

#ifndef HAVE_LSTAT
#define lstat((x),(y)) stat(x,y)
#endif 


extern char **environ;

#ifdef SYS_SIGLIST_DECLARED

extern const char * const sys_siglist[];

#else

char * sh_unix_siglist (int signum)
{

  switch (signum)
    {
#ifdef SIGHUP
    case SIGHUP: 
      return _("Hangup");
#endif
#ifdef SIGINT
    case SIGINT: 
      return _("Interrupt");
#endif
#ifdef SIGQUIT
    case SIGQUIT: 
      return _("Quit");
#endif
#ifdef SIGILL
    case SIGILL: 
      return _("Illegal instruction");
#endif
#ifdef SIGTRAP
    case SIGTRAP: 
      return _("Trace/breakpoint trap");
#endif
#ifdef SIGABRT
    case SIGABRT: 
      return _("IOT trap/Abort");
#endif
#ifdef SIGBUS
    case SIGBUS: 
      return _("Bus error");
#endif
#ifdef SIGFPE
    case SIGFPE: 
      return _("Floating point exception");
#endif
#ifdef SIGUSR1
    case SIGUSR1: 
      return _("User defined signal 1");
#endif
#ifdef SIGSEGV
    case SIGSEGV: 
      return _("Segmentation fault");
#endif
#ifdef SIGUSR2
    case SIGUSR2: 
      return _("User defined signal 2");
#endif
#ifdef SIGPIPE
    case SIGPIPE: 
      return _("Broken pipe");
#endif
#ifdef SIGALRM
    case SIGALRM: 
      return _("Alarm clock");
#endif
#ifdef SIGTERM
    case SIGTERM: 
      return _("Terminated");
#endif
#ifdef SIGSTKFLT
    case SIGSTKFLT: 
      return _("Stack fault");
#endif
#ifdef SIGCHLD
    case SIGCHLD: 
      return _("Child exited");
#endif
#ifdef SIGCONT
    case SIGCONT: 
      return _("Continued");
#endif
#ifdef SIGSTOP
    case SIGSTOP: 
      return _("Stopped");
#endif
#ifdef SIGTSTP
    case SIGTSTP: 
      return _("Stopped");
#endif
#ifdef SIGTTIN
    case SIGTTIN: 
      return _("Stopped (tty input)");
#endif
#ifdef SIGTTOU
    case SIGTTOU: 
      return _("Stopped (tty output)");
#endif
#ifdef SIGURG
    case SIGURG: 
      return _("Urgent condition");
#endif
#ifdef SIGXCPU
    case SIGXCPU: 
      return _("CPU time limit exceeded");
#endif
#ifdef SIGXFSZ
    case SIGXFSZ: 
      return _("File size limit exceeded");
#endif
#ifdef SIGVTALRM
    case SIGVTALRM: 
      return _("Virtual time alarm");
#endif
#ifdef SIGPROF
    case SIGPROF: 
      return _("Profile signal");
#endif
#ifdef SIGWINCH
    case SIGWINCH: 
      return _("Window size changed");
#endif
#ifdef SIGIO
    case SIGIO: 
      return _("Possible I/O");
#endif
#ifdef SIGPWR
    case SIGPWR: 
      return _("Power failure");
#endif
#ifdef SIGUNUSED
    case SIGUNUSED: 
      return _("Unused signal");
#endif
    }
  return _("?");
}

#endif

void sh_unix_sigexit (int signal)
{
  static char msg[62];

#if defined(SH_WITH_SERVER) && !defined(SH_WITH_CLIENT)
  strlcpy (msg, _("Exit yule on signal "), 26+1);
#else
  strlcpy (msg, _("Exit samhain on signal "), 26+1);
#endif

#ifdef SYS_SIGLIST_DECLARED
  strlcat (msg, sys_siglist[signal],     sizeof(msg));
#else
  strlcat (msg, sh_unix_siglist(signal), sizeof(msg));
#endif

  /* cleanup is done by atexit(); do not reinstall signal handler
   */
  sh_error_handle (SH_ERR_FATAL, PLACE 
		   NULL,
		   msg);

  if (signal == SIGSEGV
#ifdef SIGILL
      || signal == SIGILL
#endif
#ifdef SIGBUS
      || signal == SIGBUS
#endif
#ifdef SIGFPE
      || signal == SIGFPE
#endif
      )
    {
      memset (skey, '\0', sizeof(sh_key_t));
     _exit(EXIT_FAILURE); 
    }


  exit(EXIT_FAILURE);
}

void sh_unix_siginstall ()
{
  struct sigaction act, oldact; 

  act.sa_handler = &sh_unix_sigexit;    /* signal action           */
  sigemptyset( &act.sa_mask );          /* set an empty mask       */


  /* fallback to default action if sh_unix_sigexit fails for
   * some reason (e.g. segfault in exit)
   */
#ifdef SA_RESETHAND
  act.sa_flags = SA_RESETHAND;          /* init sa_flags           */
#else
  act.sa_flags = 0;                     /* init sa_flags           */
#endif

  /* signals that may cause termination */

#ifdef SIGHUP
  sigaction(SIGHUP,  &act, &oldact);
#endif
#ifdef SIGINT
  sigaction(SIGINT,  &act, &oldact);
#endif
#ifdef SIGQUIT
  sigaction(SIGQUIT, &act, &oldact);
#endif
#ifdef SIGILL
  sigaction(SIGILL,  &act, &oldact);
#endif
#ifdef SIGABRT
  sigaction(SIGABRT, &act, &oldact);
#endif
#ifdef SIGSEGV
  sigaction(SIGSEGV, &act, &oldact);
#endif
#ifdef SIGTRAP
  sigaction(SIGTRAP, &act, &oldact);
#endif
#ifdef SIGIOT
  sigaction(SIGIOT,  &act, &oldact);
#endif
#ifdef SIGTERM
  sigaction(SIGTERM, &act, &oldact);
#endif
#ifdef SIGBUS
  sigaction(SIGBUS,  &act, &oldact);
#endif
#ifdef SIGIO
  sigaction(SIGIO,   &act, &oldact);
#endif
#ifdef SIGPOLL
  sigaction(SIGPOLL, &act, &oldact);
#endif
#ifdef SIGXCPU
  sigaction(SIGXCPU, &act, &oldact);
#endif
#ifdef SIGPWR
  sigaction(SIGPWR,  &act, &oldact);
#endif
#ifdef SIGFPE
  sigaction(SIGFPE,  &act, &oldact);
#endif


  return;
}

/* ---------------------------------------------------------------- */
  
static   uid_t   euid;
static   uid_t   ruid;
static   gid_t   egid;
static   gid_t   rgid;

#ifdef HAVE_SETRESUID
extern       int setresuid (uid_t ruid, uid_t euid, uid_t suid);
extern       int setresgid (gid_t rgid, gid_t egid, gid_t sgid);
#endif

int sh_unix_isnotsuid()
{
  return (euid == ruid);
}

int sh_unset_suid ()
{
  int retval;

  if (ruid == euid)
    return GOOD;

#ifdef HAVE_SETRESUID
  retval = setresgid (-1, rgid, -1);
  if (retval == 0) 
    retval = setresuid (-1, ruid, -1);
#else

#ifdef  HAVE_SETEUID
  retval = setegid (rgid);
  if (retval == 0) 
    retval = seteuid (ruid);
#else

#ifdef  HAVE_SETREUID
  retval = setregid (egid, rgid);
  if (retval == 0) 
    retval = setreuid (euid, ruid);
#else
  retval = setgid (rgid);
  if (retval == 0) 
    retval = setuid (ruid);
#endif

#endif

#endif


  if (retval != 0)
    {
      retval = errno;
      fprintf (stderr, _("sh_unset_suid: Could not set UID: %s"), 
	       sh_error_message(retval));
      _exit (EXIT_FAILURE);
    }
  return GOOD;
}

int sh_set_suid ()
{
  int retval;

  if (ruid == euid)
    return GOOD;

  /* setresuid is comfortable, thus it gets priority.
   */
#ifdef  HAVE_SETRESUID
  retval = setresuid (-1, euid, -1);
  if (retval == 0) 
    retval = setresgid (-1, egid, -1);
#else 

  /* seteuid appers to be resonable well behaved.
   */
#ifdef  HAVE_SETEUID
  retval = seteuid (egid);
  if (retval == 0) 
    retval = setegid (euid);
#else

  /* on AIX, setreuid does not behave well for non-root users.
   */
#ifdef  HAVE_SETREUID
  retval = setreuid (ruid, euid);
  if (retval == 0) 
    retval = setregid (rgid, egid);
#else
  /* setuid most probably will drop root terminally.
   */
  retval = setuid (euid);
  if (retval == 0) 
    retval = setgid (egid);
#endif

#endif

#endif

  if (retval != 0)
    {
      retval = errno;
      fprintf (stderr, _("sh_set_suid: Could not set UID: %s"), 
	       sh_error_message(retval));
      _exit (EXIT_FAILURE);
    }
  return GOOD;
}

void sh_save_uids()
{
  static int stored = BAD;

  if (stored == GOOD) return;

  ruid = getuid();
  euid = geteuid();
  
  rgid = getgid();
  egid = getegid();

  if (sh_unset_suid() != GOOD)
    {
      fprintf (stderr, _("CAN'T HAPPEN.\n"));
      _exit(EXIT_FAILURE);
    }

  stored = GOOD;
  return;
}


/* drop SUID privileges
 */
int sh_unix_droppriv()
{
  sh_save_uids();
  return 0;
  /*
  setgid(getgid());
  return setuid(getuid());
  */
}

#ifdef __STDC__
char * trusterror(int errorcode)
#else
char * trusterror(errorcode)
int errorcode
#endif
{
  switch (errorcode)
    {
    case TF_BADUID: 
      return _("Owner not trustworthy");
    case TF_BADGID:
      return _("Group writeable and member not trustworthy");
    case TF_BADOTH:
      return _("World writeable");
    case TF_BADFILE:
      return _("Illegal filename");
    case TF_BADNAME:
      return _("Invalid filename (prob. too long)");
    case TF_NOROOM:
      return _("Filename too long");
    case TF_BADSTAT:
      return _("stat() failed");
    default:
      return _("Unknown error");
    }
}



/* added    Tue Feb 22 10:36:44 NFT 2000 Rainer Wichmann                */
int tf_add_trusted_user(char * c)
{
  register struct passwd *          w;
  static   int             count =  0;
  register long            pwid  = -1;

  if (count == 6)
    {
      fprintf(stderr, _("ERROR  : Maximum 7 trusted users: %s\n"), c);
      return BAD;
    }
  
  /* first check for a user name
   */
  if ((w = getpwnam(c)) != NULL && ((pwid = w->pw_uid) > 0))
    {
      rootonly[EUIDSLOT] = pwid;
      ++count;
      ++EUIDSLOT;
      return GOOD;
    }
	
  /* failed, so check for a numerical value
   */
  pwid = strtol(c, (char **)NULL, 10);
  if (pwid > 0 && pwid < 65535)
    {
      rootonly[EUIDSLOT] = pwid;
      ++count;
      ++EUIDSLOT;
      return GOOD;
    }
      
  fprintf(stderr, _("ERROR  : Invalid user: %s\n"), c);
  return BAD;
}

int tf_trust_check (char * file)
{
  char * tmp;
  char * tmp2;

  tf_euid    = euid;
  tf_path[0] = '\0';

  if (TF_YES != trustfile(file, NULL, NULL)) 
    {
      tmp  = sh_util_safe_name (file);
      if (tf_path[0] != '\0')
	{
	  tmp2  = sh_util_safe_name (tf_path);
	  sh_error_handle(SH_ERR_ERR, PLACE
			  NULL,
			  _("File <%s>: %s: %s."),
			  tmp, trusterror(tf_errno), tf_path);
	  SH_FREE(tmp2);  
	}
      else
	{
	  sh_error_handle(SH_ERR_ERR, PLACE
			  NULL,
			  _("File <%s>: %s."),
			  tmp, trusterror(tf_errno));
	}
      SH_FREE(tmp);
      return BAD;
    }
  return GOOD;
}


/* determine file type
 */
static int sh_unix_getinfo_type (struct stat * buf, 
				 ShFileType * type, char * c_mode)
{
  if      ( S_ISREG(buf->st_mode)  ) { 
    (*type)   = SH_FILE_REGULAR;
    c_mode[0] = '-';
  }
  else if ( S_ISLNK(buf->st_mode)  ) {
    (*type)   = SH_FILE_SYMLINK;
    c_mode[0] = 'l';
  }
  else if ( S_ISDIR(buf->st_mode)  ) {
    (*type)   = SH_FILE_DIRECTORY;
    c_mode[0] = 'd';
  }
  else if ( S_ISCHR(buf->st_mode)  ) {
    (*type)   = SH_FILE_CDEV;
    c_mode[0] = 'c';
  }
  else if ( S_ISBLK(buf->st_mode)  ) {
    (*type)   = SH_FILE_BDEV;
    c_mode[0] = 'b';
  }
  else if ( S_ISFIFO(buf->st_mode) ) {
    (*type)   = SH_FILE_FIFO;
    c_mode[0] = '|';
  }
  else if ( S_ISSOCK(buf->st_mode) ) {
    (*type)   = SH_FILE_SOCKET;
    c_mode[0] = 's';
  }
  else                              {
    (*type)   = SH_FILE_UNKNOWN;
    c_mode[0] = '?';
  }

  return EXIT_SUCCESS;
}


static int  sh_unix_getinfo_mode (struct stat *buf, 
				  unsigned int * mode, char * c_mode)
{
  (*mode) = buf->st_mode;

  /* make 'ls'-like string */
  
  if ( (buf->st_mode & S_IRUSR) != 0 )  c_mode[1] = 'r'; 
  if ( (buf->st_mode & S_IWUSR) != 0 )  c_mode[2] = 'w'; 
  if ( (buf->st_mode & S_IXUSR) != 0 ) {
    if ((buf->st_mode & S_ISUID) != 0 ) c_mode[3] = 's';
    else                                c_mode[3] = 'x';
  } else {
    if ((buf->st_mode & S_ISUID) != 0 ) c_mode[3] = 'S';
  }

  if ( (buf->st_mode & S_IRGRP) != 0 )  c_mode[4] = 'r'; 
  if ( (buf->st_mode & S_IWGRP) != 0 )  c_mode[5] = 'w'; 
  if ( (buf->st_mode & S_IXGRP) != 0 )  {
    if ((buf->st_mode & S_ISGID) != 0 ) c_mode[6] = 's';
    else                                c_mode[6] = 'x';
  } else {
    if ((buf->st_mode & S_ISGID) != 0 ) c_mode[6] = 'S';
  } 

  if ( (buf->st_mode & S_IROTH) != 0 )  c_mode[7] = 'r'; 
  if ( (buf->st_mode & S_IWOTH) != 0 )  c_mode[8] = 'w';
#ifdef S_ISVTX  /* not POSIX */
  if ( (buf->st_mode & S_IXOTH) != 0 )  {
    if ((buf->st_mode & S_ISVTX) != 0 ) c_mode[9] = 't';
    else                                c_mode[9] = 'x';
  } else {
    if ((buf->st_mode & S_ISVTX) != 0 ) c_mode[9] = 'T';
  }
#else
  if ( (buf->st_mode & S_IXOTH) != 0 )  c_mode[9] = 'x';
#endif 

  return EXIT_SUCCESS;
}


void sh_unix_closeall (int fd)
{
    int fdlimit = sysconf(_SC_OPEN_MAX);

    while (fd < fdlimit)
      close(fd++);
    return;
}

static void sh_unix_setlimits()
{
  struct rlimit limits;

  limits.rlim_cur = RLIM_INFINITY;
  limits.rlim_max = RLIM_INFINITY;

#ifdef RLIMIT_CPU
  setrlimit (RLIMIT_CPU,     &limits);
#endif
#ifdef RLIMIT_FSIZE
  setrlimit (RLIMIT_FSIZE,   &limits);
#endif
#ifdef RLIMIT_DATA
  setrlimit (RLIMIT_DATA,    &limits);
#endif
#ifdef RLIMIT_STACK
  setrlimit (RLIMIT_STACK,   &limits);
#endif
#ifdef RLIMIT_RSS
  setrlimit (RLIMIT_RSS,     &limits);
#endif
#ifdef RLIMIT_NPROC
  setrlimit (RLIMIT_NPROC,   &limits);
#endif
#ifdef RLIMIT_NOFILE
  setrlimit (RLIMIT_NOFILE,  &limits);
#endif
#ifdef RLIMIT_MEMLOCK
  setrlimit (RLIMIT_MEMLOCK, &limits);
#endif

  /* no core dumps
   */
  limits.rlim_cur = 0;
  limits.rlim_max = 0;
#ifdef RLIMIT_CORE
  setrlimit (RLIMIT_CORE,    &limits);
#endif

  return;
}

/* delete all environment variables
 */
static void sh_unix_zeroenv()
{
  char * c;
  char ** env = environ;

  while (*env != NULL) {
    c = strchr ((*env), '=');
    ++c;
    ++(env);
    while ((*c) != '\0') {
      (*c) = '\0';
      ++c;
    }
  }
  return;
}


static void  sh_unix_resettimer()
{
  struct itimerval this_timer;

  this_timer.it_value.tv_sec  = 0;
  this_timer.it_value.tv_usec = 0;

  setitimer(ITIMER_REAL,    &this_timer, NULL);
  setitimer(ITIMER_VIRTUAL, &this_timer, NULL);
  setitimer(ITIMER_PROF,    &this_timer, NULL);

  return;
}

static void  sh_unix_resetsignals()
{
  int  sig_num;
#ifdef NSIG
  int  max_sig = NSIG;  /* FIXME is NSIG the maximum number ? */
#else
  int  max_sig = 255;
#endif
  int  test;
  struct sigaction act, oldact;
  int  status;
  
  act.sa_handler = SIG_DFL;         /* signal action           */
  sigemptyset( &act.sa_mask );      /* set an empty mask       */
  act.sa_flags = 0;                 /* init sa_flags           */

  for (sig_num = 1; sig_num <= max_sig; ++sig_num) 
    {
      test = sigaction(sig_num,  &act, &oldact);
      if ((test == -1) && (errno != EINVAL)) 
	{
	  status = errno;
	  sh_error_handle (SH_ERR_WARN, PLACE
			   _("Initialization failed"),
			   _("Failed reset of signal %d: %s"),
			   sig_num,
			   sh_error_message (status));
	}
    }

  return;
}

/* get the local hostname (FQDN)
 */
#include <netdb.h>
#include <sys/socket.h> 
#include <arpa/inet.h>
#ifdef HAVE_UNAME
#include <sys/utsname.h>
void sh_unix_localhost()
{
  struct utsname   buf;
  struct hostent * he1;
  int              i;
  int              ddot = 0;
  int              len;


  (void) uname (&buf);
  he1 = gethostbyname(buf.nodename);
  strlcpy (sh.localhost, he1->h_name, SH_MINIBUF+1);

  /* check whether it looks like a FQDN
   */
  len = strlen(sh.localhost);
  for (i = 0; i < len; ++i) 
    if (sh.localhost[i] == '.') ++ddot; 
  if (ddot == 0) 
    strlcpy (sh.localhost, 
	     inet_ntoa (*(struct in_addr *) he1->h_addr), 
	     SH_MINIBUF+1); 

  return;
}
#else
void sh_unix_localhost()
{
  struct hostent * he1;

  (void) gethostname (sh.localhost, SH_MINIBUF);
  sh.localhost[SH_MINIBUF] = '\0';
  he1 = gethostbyname(sh.localhost);
  strlcpy (sh.localhost, he1->h_name, SH_MINIBUF+1);

  /* check whether it looks like a FQDN
   */
  len = strlen(sh.localhost);
  for (i = 0; i < len; ++i) 
    if (sh.localhost[i] == '.') ++ddot; 
  if (ddot == 0) 
    strlcpy (sh.localhost, 
	     inet_ntoa (*(struct in_addr *) he1->h_addr), 
	     SH_MINIBUF+1); 

  return;
}
#endif


void sh_unix_mlock()
{
#if defined(HAVE_MLOCK) && !defined(HAVE_BROKEN_MLOCK)
  if (skey->mlock_failed == GOOD)
    {
      sh_set_suid ();
      if ( (-1) == mlock( (char *) skey, sizeof (sh_key_t)) ) 
	{
	  skey->mlock_failed = BAD;
	}
      sh_unset_suid ();
    }
#else
  if (skey->mlock_failed == GOOD)
    {
      skey->mlock_failed = BAD;
    }
#endif
}
  
/* daemon mode 
 */
int sh_unix_setdeamon(char * dummy)
{
  sh.modus = 1;
  return 0;
}

int sh_unix_init(int goDaemon)
{
  int status;

  /* fork twice, exit the parent process
   */
  if (goDaemon == 1) {
    
    switch (fork()) {
    case 0:  break;      /* child process continues */
    case -1: return -1;  /* error                   */
    default: _exit(0);   /* parent preocess exits   */
    }

    /* Child processes do not inherit page locks across a fork.
     * Error in next fork would return in this (?) thread of execution.
     */
    sh_unix_mlock();
    setsid();            /* should not fail         */

    switch (fork()) {
    case 0:  break;      /* child process continues */
    case -1: return -1;  /* error                   */
    default: _exit(0);   /* parent preocess exits   */
    }

    /* Child processes do not inherit page locks across a fork.
     */
    sh_unix_mlock();

  } else {
    setsid();            /* should not fail         */
  } 

  /* set working directory   
   */
  status = chdir("/");
  if ( (-1) == status ) 
    {
      sh_error_handle (SH_ERR_WARN, PLACE
		       _("Chdir failed"),
		       _("Failed chdir(): %s"),
		       sh_error_message (status));
    }

  /* reset timers 
   */
  sh_unix_resettimer();

  /* signal handlers 
   */
  sh_unix_resetsignals();
  sh_unix_siginstall ();

  /* set resource limits to maximum 
   */
  sh_unix_setlimits();

  /* set file creation mask 
   */
  (void) umask (0); /* should not fail */

  /* zero out the environment (like PATH='\0')  
   */
  sh_unix_zeroenv();

  /* close all file descriptors 
   */
  if (goDaemon == 1) sh_unix_closeall (0);

  /* open first three streams 
   */
  if (goDaemon == 1) {
    open(_("/dev/null"), O_RDWR);
    dup(0); 
    dup(0);
  }

  return 0;
}

unsigned long sh_unix_longtime ()
{
  time_t        time_now;

  (void) time (&time_now);
  return (unsigned long)time_now;
} 

int sh_unix_settimeserver (char * address)
{
  unsigned long  numcop;

  if (address != NULL) 
    {
      numcop = strlcpy (sh.timeserver, address, 256+1);
      if (numcop >= 257) 
	{
	  fprintf(stderr, 
		  _("ERROR : Mail address too long (max=256): %s"), 
		  address);
	  return BAD;
	}
      return GOOD;
    }
  return BAD;
}


#ifdef HAVE_NTIME
#define UNIXEPOCH 2208988800UL  /* difference between Unix time and net time 
                                 * The UNIX EPOCH starts in 1970.
                                 */
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <ctype.h>
#endif

/* Timeserver service.               */
/* define is missing on HP-UX 10.20  */
#ifndef IPPORT_TIMESERVER 
#define IPPORT_TIMESERVER 37 
#endif


char * sh_unix_time (unsigned long thetime)
{
#ifdef HAVE_STRFTIME
  int           status;
#endif
  time_t        time_now;
  struct tm   * time_ptr;
  static char   AsciiTime[81];                       /* local time   */
  static char   deftime[] = N_("[00-00-0000 00:00:00]"); /* default time */
  int    fail = 0;              /* no net time                              */


#ifdef HAVE_NTIME
  struct in_addr       haddr;   /* host address from numeric                */
                                /* host details returned by the DNS         */
  struct hostent *host_entry=NULL;   
  static struct sockaddr_in sin;/* socket to the remote host                */
  int    fd;                    /* network file descriptor                  */
  u_char net_time[4];           /* remote time in network format            */
  static int failerr = 0;       /* no net time                              */
  fd_set fds;
  struct timeval tv;
  int    testflag;
  char   errmsg[256];

  static int cached = 0;
#endif
  
  if (thetime == 0) {

#ifdef HAVE_NTIME
    if (strncmp(_("NULL"), sh.timeserver, 4) == 0) {
      fail = 1;
#endif

      (void) time (&time_now);

#ifdef HAVE_NTIME
    } else {  /* have a timeserver address */

      /* ---- Get Time from Timeserver ---- */

      /* only use gethostbyname() if neccessary
       */
      testflag = GOOD;
#ifdef HAVE_INET_ATON
      if ( (cached == 0) &&  (0 == inet_aton(sh.timeserver, &haddr)) )
#else
      if ( (cached == 0) && 
	   ((unsigned long)-1  == (haddr.s_addr = inet_addr(sh.timeserver))) )
#endif
	{
	  testflag = BAD;
	  if ((host_entry = gethostbyname(sh.timeserver)) == NULL) 
	    {
	      strlcpy(errmsg, _("gethostbyname(%s) failed"), 256);
	      fail = 1;
	    }
	  else
	    {
	      sin.sin_family = AF_INET;
	      sin.sin_port   = htons (IPPORT_TIMESERVER);
	      sin.sin_addr   = *(struct in_addr *) host_entry->h_addr;
	      cached         = 1;
	    }
	}
      else if (cached == 0)
	{
	  sin.sin_family = AF_INET;
	  sin.sin_port   = htons (IPPORT_TIMESERVER);
	  sin.sin_addr   = haddr;
	  cached         = 1;
	}
	


      if (fail != 1) 
	{ 
	  fd = socket(AF_INET, SOCK_STREAM, 0);
	  if (fd < 0) {
	    fail = 1;
	    strlcpy(errmsg, _("Timeserver %s - socket(): "), 256);
	    strlcat(errmsg, sh_error_message (errno), 256);
	  }
	}
      else
	{
	  fd = -1;
	}
      
      if (fail != 1) {

	if (connect(fd, (struct sockaddr *) &sin, sizeof(sin)) < 0) 
	  {
	    strlcpy(errmsg, _("Timeserver %s - connect(): "), 256);
	    strlcat(errmsg, sh_error_message (errno), 256);
	    close(fd);
	    fail = 1; 
	  }
	else 
	  {
	    FD_ZERO(&fds);
	    FD_SET(fd, &fds);
	    
	    tv.tv_sec  = 2;
	    tv.tv_usec = 0;

	    if (select(fd+1, &fds, NULL, NULL, &tv) > 0) {
	      if (read(fd, net_time, 4) != 4) {     
		strlcpy(errmsg, _("Timeserver %s - read(): "), 256);
		strlcat(errmsg, sh_error_message (errno), 256);
		fail = 1;
	      }
	    } else {
		strlcpy(errmsg, _("Timeserver %s - timeout expired"),   256);
		fail = 1;
	    } 
	    close(fd);
	  }

	if (fail == 0) 
	  { 
	    time_now = ntohl(* (long *) net_time) - UNIXEPOCH;
	    if (failerr == 1) {
	      failerr = 0;
	      sh_error_handle (SH_ERR_ERR, PLACE
			       NULL,
			       _("Connection to timeserver %s "\
				 "re-established."),
			       sh.timeserver);
	    } 
	  } 
      }

      if (fail == 1)
	{
	  sin.sin_family = 0;

	  (void) time (&time_now);
	  if (failerr == 0)
	    {
	      failerr = 1;
	      sh_error_handle (SH_ERR_ERR, PLACE
			       _("No connection to timeserver %s."),
			       errmsg,
			       sh.timeserver);
	    }
	}
	
    }
#endif
 
  } else {
    time_now = thetime;
  }

  if (time_now == (-1) ) { 
    if (sh.modus == 0)
      fprintf(stderr, _("Cannot determine time.\n"));
    return _(deftime);
  }

  time_ptr   = localtime (&time_now);

  if (time_ptr != NULL) {
#ifdef HAVE_STRFTIME
    if (fail == 1)
      status = strftime (AsciiTime, 80, _("(%d-%m-%Y %H:%M:%S)"), time_ptr);
    else
      status = strftime (AsciiTime, 80, _("[%d-%m-%Y %H:%M:%S]"), time_ptr);
    if ( (status == 0) || (status == 80) ) {
      if (sh.modus == 0)
	fprintf(stderr, _("Cannot determine time.\n"));
      return _(deftime);
    }
    return &AsciiTime[0];
#else
    if (fail == 1)
      printf (AsciiTime, _("(%2d-%2d-%4d %2d:%2d:%2d)"), 
	      time_ptr->tm_mday, 1+time_ptr->tm_mon, 1900+time_ptr->tm_year,
	      time_ptr->tm_hour, time_ptr->tm_min, time_ptr->tm_sec);
    else
      printf (AsciiTime, _("[%2d-%2d-%4d %2d:%2d:%2d]"), 
	      time_ptr->tm_mday, 1+time_ptr->tm_mon, 1900+time_ptr->tm_year,
	      time_ptr->tm_hour, time_ptr->tm_min, time_ptr->tm_sec);
    return &AsciiTime[0];
#endif
  } else {
    if (sh.modus == 0)
      fprintf(stderr, _("Cannot determine time.\n"));
    return _(deftime);
  }

  /* notreached */
}

/**************************************
char * sh_unix_gethome (char * name)
{
  struct passwd * tempres;

  tempres = getpwnam(name);
  if (tempres == NULL) {
    sh_error_handle (SH_ERR_ERR, PLACE
		     _("No such user: %s"),
		     _("Failed getpwnam() on %s"),
		     name);
    return NULL;
  }
  return tempres->pw_dir;
}

uid_t sh_unix_getuid (char * name)
{
  struct passwd * tempres;
  uid_t  theuid;

  tempres = getpwnam(name);
  if (tempres == NULL) {
    sh_error_handle (SH_ERR_ERR, PLACE
		     _("No such user: %s"),
		     _("Failed getpwnam() on %s"),
		     name);
    return -1;
  }

  theuid = tempres->pw_uid;
  return theuid;
}
****************************************/

int sh_unix_getUser ()
{
  struct passwd * tempres;

  tempres = getpwuid(geteuid());

  if (tempres == NULL) return BAD;

  if (tempres->pw_name != NULL) {
    if (strlen(tempres->pw_name) > SH_MINIBUF) {
      sh_error_handle (SH_ERR_ERR, PLACE
		       _("Username too long"), 
		       _("Username (%s) exceeds limit"), 
		       tempres->pw_name );
      return BAD;
    } else {
      strlcpy ( sh.user, tempres->pw_name, SH_MINIBUF+1);
    }
  } else {
    sh_error_handle (SH_ERR_ERR, PLACE
		     _("Cannot find out who you are"), 
		     _("Password field pw_user is NULL") );
    sh.user[0] = '\0';
    return BAD;
  }

  if (tempres->pw_gecos != NULL) {
    if (strlen(tempres->pw_gecos) > SH_MINIBUF) {
      sh_error_handle (SH_ERR_ERR, PLACE
		       _("Real username too long"), 
		       _("Gecos field (%s) exceeds limit"), 
		       tempres->pw_gecos );
      return BAD;
    } else {
      strlcpy ( sh.real_user, tempres->pw_gecos, SH_MINIBUF+1);
    }
  } else {
    sh_error_handle (SH_ERR_WARN, PLACE
		     _("Cannot determine real username"), 
		     _("Password field pw_gecos is NULL") );
    sh.real_user[0] = '\0';
  }


  if (tempres->pw_dir != NULL) {
    if (strlen(tempres->pw_dir) > SH_BUFSIZE) {
      sh_error_handle (SH_ERR_ERR, PLACE
		       _("Home directory too long"), 
		       _("Home directory path too long (%s)"), 
		       tempres->pw_dir );
      return BAD;
    } else {
      strlcpy ( sh.home, tempres->pw_dir, SH_BUFSIZE+1);
      return GOOD;
    }
  } else {
    sh_error_handle (SH_ERR_ERR, PLACE
		     _("Cannot determine home directory"), 
		     _("Password field pw_dir is NULL") );
    sh.home[0] = '\0';
    return BAD;
  }

  /* notreached */
}


char *  sh_unix_getUIDname (int level, uid_t uid)
{
  struct passwd * tempres;

  tempres = getpwuid(uid);

  if (tempres == NULL) {
    sh_error_handle (level, PLACE
		     _("Cannot find user for UID %ld"), 
		     _("Password file entry is NULL for UID %ld"),
		     (long) uid);
    return NULL;
  }

  if (tempres->pw_name != NULL) {
      return tempres->pw_name;
  } else {
    sh_error_handle (level, PLACE
		     _("Cannot find user for UID %ld"), 
		     _("Password field pw_user is NULL for UID %ld"),
		     (long) uid);
    return NULL;
  }
}

char *  sh_unix_getGIDname (int level, gid_t gid)
{
  struct group * tempres;

  tempres = getgrgid(gid);

  if (tempres == NULL) {
    sh_error_handle (level, PLACE
		     _("Cannot find group for GID %ld"), 
		     _("Group file entry is NULL for GID %ld"),
		     (long) gid);
    return NULL;
  }

  if (tempres->gr_name != NULL) {
      return tempres->gr_name;
  } else {
    sh_error_handle (level, PLACE
		     _("Cannot find group for GID %ld"), 
		     _("Group field pw_user is NULL for GID %ld"),
		     (long) gid);
    return NULL;
  }
}


int  sh_unix_fileopen (char * filename, char mode, 
		       ShOpenType IsType, int level, int withpriv)
{
  struct stat   lbuf;
  struct stat   buf;
  int           lstat_return;
  int           stat_return;
  struct passwd *filepwd;
  int           fd;
  char        * tmp;
  char        * tmp2;
  char        * tmpdir;
  char          message[81];
  static  int   firstaccess = 1;

  (void) umask (0);
  errno = 0;

  lstat_return = lstat(filename, &lbuf);

  if (lstat_return == -1 && errno != ENOENT) {
    lstat_return = errno;
    if (IsType == SH_ISLOG)  sh_error_logoff();
    tmp = sh_util_safe_name (filename);
    sh_error_handle (level, PLACE
		     _("Cannot check file: %s"),
		     _("Cannot lstat() file: %s: %s"),
		     tmp, 
		     sh_error_message (lstat_return));
    SH_FREE(tmp);
    return (-1);
  } 
  
  if (errno == ENOENT) 
    lstat_return = ENOENT;
  
  switch (mode) 
    {
    case 'w':
      if (YESPRIV == withpriv) 
	tf_euid = euid;  /* PRIV */
      else 
	tf_euid = ruid;
      tmpdir = sh_util_basename(filename);

      tf_path[0] = '\0';

      if (TF_YES != trustfile (tmpdir, NULL, NULL)) 
	{
	  if (IsType == SH_ISLOG)  sh_error_logoff();
	  tmp  = sh_util_safe_name (filename);
	  if (tf_path[0] != '\0')
	    {
	      tmp2 = sh_util_safe_name (tf_path);
	      sh_error_handle (SH_ERR_FATAL, PLACE
			       NULL,
			       _("Write to <%s>: %s: %s"),
			       tmp, trusterror(tf_errno), tmp2);
	      SH_FREE(tmp2);
	    }
	  else
	    {
	      sh_error_handle (SH_ERR_FATAL, PLACE
			       NULL,
			       _("Write to <%s>: %s"),
			       tmp, trusterror(tf_errno));
	    }
	  SH_FREE(tmp);
	  SH_FREE(tmpdir);
	  exit (EXIT_FAILURE);
	}
      SH_FREE(tmpdir);

      /* O_EXECL fixes race condition between lstat() and open() 
       * -- might fail on NFS, so we do an fstat() later --
       */
      if (YESPRIV == withpriv) sh_set_suid();
      if (lstat_return == ENOENT)
	fd = open (filename, O_WRONLY|O_CREAT|O_EXCL, S_IWUSR|S_IRUSR|S_IRGRP);
      else
	fd = open (filename, O_WRONLY|O_CREAT, S_IWUSR|S_IRUSR|S_IRGRP);
      if (YESPRIV == withpriv) sh_unset_suid();
      break;

    case 'r':
      if (YESPRIV == withpriv) sh_set_suid();
      fd = open (filename, O_RDONLY);
      if (YESPRIV == withpriv) sh_unset_suid();
      break;

    default:
      return (-1);
      break;
    } 

  
  if (fd == -1) 
    {
      stat_return = errno;
      if (IsType == SH_ISLOG)  sh_error_logoff();
      tmp = sh_util_safe_name (filename);
      sh_error_handle (level, PLACE
		       _("Cannot open file: %s"),
		       _("Cannot open file: %s: %s"),
		       tmp,
		       sh_error_message (stat_return));
      SH_FREE(tmp);
      return (-1);
    } 
    
  /* by default, append to file
   */
  if (mode == 'w') 
    lseek (fd, 0L, SEEK_END);

  stat_return = fstat(fd, &buf);

  if ( (stat_return == -1) ) 
    {
      /* does not exist / permission denied          */
      stat_return = errno;
      if (IsType == SH_ISLOG)  sh_error_logoff();
      tmp = sh_util_safe_name (filename);
      sh_error_handle (level, PLACE
		       _("Cannot access file: %s"),
		       _("Cannot stat() file: %s: %s"),
		       tmp, 
		       sh_error_message (stat_return));
      SH_FREE(tmp);
      return (-1);
    } 
  else 
    {
      /* is it what we want ? If it had not existed before, lstat()
       * will have failed with ENOENT
       */
      if (lstat_return != ENOENT && buf.st_ino != lbuf.st_ino) 
	{
	  if (IsType == SH_ISLOG)  sh_error_logoff();
	  tmp = sh_util_safe_name (filename);
	  sh_error_handle (level, PLACE
			   _("Do not access because possible link: %s"),
			   _("lstat() st_ino != fstat() st_ino: %s"),
			   tmp);
	  SH_FREE(tmp);
	  return (-1);
	}

      /* exists, check permissions and ownership */

      stat_return = 0;

      if (mode == 'w') 
	{
	  /* check on lock
	   */
	  if (firstaccess == 1 && IsType == SH_ISLOG) 
	    {
	      if (buf.st_nlink > 1) 
		{
		  sh_error_logoff();
		  tmp = sh_util_safe_name (filename);
		  sh_error_handle (level, PLACE
				   _("Logfile %s is locked"),
				   _("Logfile %s is locked (%d hardlinks)"),
				   tmp, 
				   buf.st_nlink);
		  SH_FREE(tmp);
		  return (-1);
		} 
	      else if (IsType == SH_ISLOG) 
		{
		  sh_unix_lock (filename, DEFAULT_ERRLOCK);
		  firstaccess = 0;
		}
	    }
    
	  if ( buf.st_uid != geteuid() ) 
	    {  
	      /* we must be owner to write */
	      /* find the owner            */
	      stat_return = -1;
	      filepwd = getpwuid(buf.st_uid);
	      strlcpy(message, _("File owned by user: "), 81);
	      strlcat(message, filepwd->pw_name, 81 );
	      strlcat(message, _(": %s"), 81 );
	    }

	  if ( !S_ISREG(buf.st_mode) && 
	       (IsType == SH_ISFILE || IsType == SH_ISLOG)) 
	    {   
	      /* not a regular file */
	      stat_return = -1;
	      strlcpy(message, _("File is not a regular file: %s"), 81);
	    }

	  if ( (buf.st_mode & S_IWOTH)  && (IsType == SH_ISLOG) ) 
	    {
	      stat_return = -1;
	      strlcpy(message, _("Logfile is world writeable: %s"), 81);
	    }

	} 
      else if (mode == 'r') 
	{

	  if ( (IsType == SH_ISDIR) && (!S_ISDIR(buf.st_mode)) ) 
	    {   
	      /* is it a directory ? */
	      stat_return = -1;
	      strlcpy(message, _("File is not a directory: %s"), 81);
	    }

	  if (IsType == SH_ISDATA) 
	    {
	      if ( (buf.st_mode & S_IWOTH)  && (IsType == SH_ISDATA) ) 
		{
		  stat_return = -1;
		  strlcpy(message, _("Datafile is world writeable: %s"), 81);
		}
	      if ( !S_ISREG(buf.st_mode) ) 
		{   
		  /* not a regular file */
		  stat_return = -1;
		  strlcpy(message, _("File is not a regular file: %s"), 81);
		}
	    }
	}

    if (stat_return == -1) 
      {
	if (IsType == SH_ISLOG)  sh_error_logoff();
	tmp = sh_util_safe_name (filename);
	sh_error_handle (level, PLACE
			 NULL,
			 message,
			 tmp);
	SH_FREE(tmp);
	sh_unix_fileclose (fd);
	return (-1);
      }
    }

#ifdef MEM_DEBUG
  ++sh.openfiles;
  sh_mem_openf(filename, fd);
#endif
  return fd;
}
  
int sh_unix_filewrite (int fd, void * msg, long nbytes, 
		       ShOpenType IsType, ShDataType IsData)
{
  int  stat_return = 0;
  int  err_store   = 0;
  int  err_sync    = 0;
  long bytewritten;
  long bytecount;

  int  level = SH_ERR_ERR;

  if (IsType == SH_ISLOG)  level = SH_ERR_FATAL;
  if (IsType == SH_ISDATA) level = SH_ERR_FATAL;
  if (IsType == SH_ISFILE) level = SH_ERR_ERR;
  if (IsType == SH_ISDIR)  level = SH_ERR_ERR;

  /* write message
   */
  bytecount    = 0;
  bytewritten  = 0;
  while (bytecount < nbytes) 
    {    
      if ((bytewritten = write (fd, msg, nbytes-bytecount)) > 0) 
	{
	  bytecount += bytewritten;
	  msg       += bytewritten;    /* move buffer pointer forward */
	}
      else if (bytewritten < 0)
	{
	  if ( errno == EINTR || errno == EAGAIN) /* try again */
	      continue;
	  else 
	    {
	      err_store = errno;
	      if (IsType == SH_ISLOG)  sh_error_logoff();
	      sh_error_handle (level, PLACE
			       _("A write error occurred"),
			       _("write() error: %s"), 
			       sh_error_message (err_store));
	      if (IsType == SH_ISLOG)  sh_error_logrestore();
	      return (EXIT_FAILURE);
	    }
	}
    }

  if (IsData == SH_DATA_LINE) 
    {
      stat_return = 0;
      while (stat_return < 1)
	{
	  stat_return = write (fd, "\n", 1);
	  if (stat_return < 0)
	    {
	      if (errno == EINTR || errno == EAGAIN) /* try again */
		continue;
	      else
		break;
	    }
	}
    }

  if (IsType == SH_ISLOG) err_sync = fsync (fd);
  if (err_sync == (-1)) {
    err_sync = errno;
    sh_error_logoff();
    sh_error_handle (level, PLACE
		     _("Could not syncronize logfile"),
		     _("fsync() error: %s"), 
		     sh_error_message (err_sync));
    sh_error_logrestore();
  }

  return (EXIT_SUCCESS);
}

int sh_unix_getline (int fd, char * line, int sizeofline)
{
  register int  count;
  register int  n = 0;
  char          c;

  if (sizeofline < 2) {
    line[n] = '\0';
    return 0;
  }

  while (n < sizeofline) {

    count = sh_unix_fileread (fd, &c, 1);

    /* end of file
     */
    if (count < 1) {
      line[n] = '\0';
      n = -1;
      break;
    } 

    if (c != '\0' && c != '\n') {
      line[n] = c;
      ++n;
    } else if (c == '\n') {
      if (n > 0) {
	line[n] = '\0';
	break;
      } else {
	line[n] = '\n'; /* get newline only if only char on line */
	++n;
	line[n] = '\0';
	break;
      }
    } else {
      line[n] = '\0';
      break;
    }

  }

  line[sizeofline] = '\0';  /* make sure line is terminated */
  return n;
}



int sh_unix_fileread (int fd, void * buf, size_t count)
{
  int store;
  int byteread = 0;
  int bytes    = 0;

  do {
    byteread = read (fd, buf, count);
    bytes += byteread; count -= byteread;  
    if (count == 0) break;
    buf += byteread;
  } while ( byteread > 0 || 
	    ( byteread < 0 && (errno == EINTR || errno == EAGAIN)) 
	    );

  if (byteread == (-1)) {
      store = errno;
      sh_error_handle (SH_ERR_ERR, PLACE 
		       _("Error reading file"),
		       _("Error reading file: %s"),
		       sh_error_message(store)
		       );
      return -1;
  }

  return bytes;
}

void sh_unix_fileclose (int fd)
{
  /* FIXME may fail on 'write-behind' systems (NFS) */
#ifdef MEM_DEBUG
  if (close (fd) != (-1)) {
    --sh.openfiles;
    return sh_mem_closef(fd);
  }
#else
  close (fd);
#endif
  return;
}


/* obtain file info
 */
int sh_unix_getinfo (int level, 
		     char * filename, file_type * theFile)
{
  struct stat   buf;
  struct stat   lbuf;
  int           stat_return;
  register int  i;
  ShFileType    type;
  unsigned int  mode;
  char        * name;
  char        * tmp;
  char        * tmp2;
  char        * linknamebuf;
  int           linksize;
  

  /* stat the file
   */
  stat_return = lstat (theFile->fullpath, &buf);/* FIXME lstat is not POSIX */

  /* check for error
   */
  if (stat_return != 0) {
    stat_return = errno;
    tmp2 = sh_util_safe_name (theFile->fullpath);
    sh_error_handle (level, PLACE
		     _("No file status: %s"),
		     _("No file status: %s: lstat: %s"),
		     tmp2,
		     sh_error_message (stat_return)
		     );
    SH_FREE(tmp2);
    return EXIT_FAILURE;
  }

  for (i = 0; i < 10; ++i) theFile->c_mode[i] = '-';
  theFile->c_mode[10] = '\0';
  for (i = 0; i < 11; ++i) theFile->link_c_mode[i] = '-';
  theFile->link_c_mode[10] = '\0';


  /* determine file type
   */
  sh_unix_getinfo_type (&buf, &type, theFile->c_mode);
  theFile->type = type;

  /* determine permissions
   */
  sh_unix_getinfo_mode (&buf, &mode, theFile->c_mode);
  theFile->mode = mode;

  /* trivia
   */
  theFile->dev       = (unsigned long) buf.st_dev;
  theFile->ino       = (unsigned long) buf.st_ino;
  theFile->mode      = (unsigned long) buf.st_mode;
  theFile->hardlinks = (unsigned long) buf.st_nlink;
  theFile->owner     = (long)          buf.st_uid;  
  theFile->group     = (long)          buf.st_gid;  
  theFile->rdev      = (unsigned long) buf.st_rdev;
  theFile->size      = (unsigned long) buf.st_size;
  theFile->blksize   = (unsigned long) buf.st_blksize;
  theFile->blocks    = (unsigned long) buf.st_blocks;
  theFile->atime     = (unsigned long) buf.st_atime;
  theFile->mtime     = (unsigned long) buf.st_mtime;
  theFile->ctime     = (unsigned long) buf.st_ctime;


  /* owner and group
   */

  if ( (name = sh_unix_getGIDname(level, buf.st_gid)) != NULL) {
    strlcpy (theFile->c_group, name, GROUP_MAX+1);
  } else {
    tmp2 = sh_util_safe_name (theFile->fullpath);
    sh_error_handle (level, PLACE
		     NULL,
		     _("No group name for file: %s"),
		     tmp2);
    SH_FREE(tmp2);
    theFile->c_group[0] = '\0';
  }

  
  if ( (name = sh_unix_getUIDname(level, buf.st_uid)) != NULL) {
    strlcpy (theFile->c_owner, name, OWNER_MAX+1);
  } else {
    tmp2 = sh_util_safe_name (theFile->fullpath);
    sh_error_handle (level, PLACE
		     NULL,
		     _("No owner name for file: %s"),
		     tmp2);
    SH_FREE(tmp2);
    theFile->c_owner[0] = '\0';
  }

  /* output the file
   */
  tmp2 = sh_util_safe_name (filename);
  sh_error_handle (SH_ERR_ALL, PLACE
		   NULL,
		   _("%10s %2d %8s %8s %14ld %21s %s"), 
		   theFile->c_mode,
		   theFile->hardlinks,
		   theFile->c_owner,
		   theFile->c_group,
		   theFile->size,
		   sh_unix_time (theFile->mtime),
		   tmp2);
  SH_FREE(tmp2);

  /* check for links
   */
  if (theFile->c_mode[0] == 'l') {

    linknamebuf = SH_ALLOC(PATH_MAX + 2);

    linksize    = readlink (theFile->fullpath, linknamebuf, PATH_MAX);
    if (linksize < PATH_MAX) linknamebuf[linksize] = '\0';
    else linknamebuf[PATH_MAX] = '\0';

    if (linksize < 0) {
      linksize = errno;
      tmp2 = sh_util_safe_name (theFile->fullpath);
      sh_error_handle (level, PLACE
		       _("No link found for %s"),
		       _("Error reading link value for file %s: %s"),
		       tmp2,
		       sh_error_message (linksize));
      SH_FREE(tmp2);
      SH_FREE(linknamebuf);
      return EXIT_FAILURE;
    }
    
    if (linknamebuf[0] == '/') {
      strlcpy (theFile->linkpath, linknamebuf, PATH_MAX+1);
    } else {
      tmp = sh_util_basename(theFile->fullpath);
      strlcpy (theFile->linkpath, 
	       tmp, 
	       PATH_MAX+1);
      SH_FREE(tmp);
      strlcat (theFile->linkpath, 
	       "/", PATH_MAX+1);
      strlcat (theFile->linkpath, 
	       linknamebuf, 
	       PATH_MAX+1);
    }

    /* stat the link
     */
    stat_return = lstat (theFile->linkpath, &lbuf); 

    /* check for error
     */
    if (stat_return != 0) { 
      stat_return = errno;
      tmp  = sh_util_safe_name (theFile->fullpath);
      tmp2 = sh_util_safe_name (theFile->linkpath);
      if (errno != ENOENT) 
	sh_error_handle (level, PLACE
			 _("Could not check linked file %s"),
			 _("Could not check linked file %s: %s"),
			 tmp2,
			 sh_error_message (stat_return)
			 );
      else /* a dangling link -- everybody seems to have plenty of them */
	sh_error_handle (SH_ERR_INFO, PLACE
			 _("Dangling link %s -> %s"),
			 _("Dangling link %s -> %s: %s"),
			 tmp,
			 tmp2,
			 sh_error_message (stat_return)
			 );
      theFile->linkisok = BAD;
      SH_FREE(tmp);
      SH_FREE(tmp2);
      SH_FREE(linknamebuf);
      return EXIT_FAILURE;
    }

    theFile->linkisok = GOOD;
      

    /* determine file type
     */
    sh_unix_getinfo_type (&lbuf, &type, theFile->link_c_mode);
    theFile->type = type;
  
    /* determine permissions
     */
    sh_unix_getinfo_mode (&lbuf, &mode, theFile->link_c_mode);
    theFile->linkmode = mode;

    /* output the link
     */
    if (theFile->linkisok == GOOD) {
      tmp2 = sh_util_safe_name (linknamebuf);      
      sh_error_handle (SH_ERR_ALL, PLACE
		       NULL,
		       _("   >>>  %10s  %s"),
		       theFile->link_c_mode,
		       tmp2);
      SH_FREE(tmp2);
    }

    SH_FREE(linknamebuf);

  }
  
  return EXIT_SUCCESS;
}


void sh_unix_unlock()
{
  struct stat buf;
  int         error = 0;
  
  /* logfile is not locked to us
   */
  if (sh.locked == BAD) 
    return;

  /* check whether the directory is secure
   */
  if (BAD == tf_trust_check (DEFAULT_ERRLOCK))
    return;

  /* delete the lock file if more than one hardlink exists */
  if (0 == lstat (DEFAULT_ERRLOCK, &buf)) {
    if (buf.st_nlink > 1) {
      sh_set_suid(); 
      unlink (DEFAULT_ERRLOCK);
      sh_unset_suid(); 
      sh.locked = BAD; /* not locked anymore */
    } else {
      sh_error_handle (SH_ERR_ERR, PLACE
		       _("Error deleting lock file %s"),
		       _("Error deleting lock file %s: st_nlink < 2"),
		       DEFAULT_ERRLOCK);
    }
  } else {
    error = errno;
  }


  if (error != 0 && error != ENOENT) 
    sh_error_handle (SH_ERR_ERR, PLACE
		     _("Error deleting lock file %s"),
		     _("Error deleting lock file %s: %s"),
		     DEFAULT_ERRLOCK,
		     sh_error_message(error));
  
  return;
}

void sh_unix_lock (char * s, char * lock)
{
  struct stat buf;
  int errnum;

  if (s == NULL && lock == NULL) return;

  if (0 == lstat (lock, &buf)) {
    fprintf(stderr, _("Startup error: lockfile %s already exists\n"), lock);
    sh.locked       = BAD;
    exit (EXIT_FAILURE);
  }

  /* If newpath exists, link() will not overwrite it. If someone slips
   * in a file, it will only be deleted on unlocking if N(hardlinks) > 1. 
   * Should not harm (?).
   */
  sh_set_suid(); 
  errnum = link (s, lock);
  sh_unset_suid(); 

  if (0 != lstat (lock, &buf)) {
    fprintf(stderr, _("Startup error: error creating lockfile %s: %s\n"), 
	    lock, sh_error_message (errnum));
    sh.locked       = BAD;
    exit (EXIT_FAILURE);
  }

  sh.locked       = GOOD; /* logfile is locked to us */
  return;
}

/***************************************
int sh_unix_test_proc()
{
#ifdef STATFS_WORKS
  struct statfs       buf;
  int                 error;

  if (0 != statfs("/proc", &buf)) {
    error = errno;
    sh_error_handle(SH_ERR_ERR, PLACE
		      _("Cannot access /proc"),
		      _("Failed statfs() on /proc: %s"),
		      sh_error_message(error));
    return BAD;
  } else {
#ifdef SH_PROC_MAGIC
    if (buf.f_type != SH_PROC_MAGIC) {
      sh_error_handle(SH_ERR_ERR, PLACE
		      NULL,
		      _("Not a proc filesystem: /proc"));
      return BAD;
    }
#endif
    return GOOD;
  }
#else
  return GOOD;
#endif
}
************************************************/

/* test whether file exists
 */
int sh_unix_file_exists(char * file_name)
{
  
  struct stat buf;
  if (stat(file_name, &buf) == -1)
    return BAD;
  else if ( S_ISCHR(buf.st_mode) &&  0 != (S_IROTH & buf.st_mode)) 
    return GOOD;
  else 
    return BAD;
}

/************************************************/
/************************************************/
/****   Stealth Utilities                    ****/
/************************************************/
/************************************************/
#ifdef SH_STEALTH

void sh_unix_xor_code (char * str, int len)
{
  register int i;

  for (i = 0; i < len; ++i) str[i] ^= (char) XOR_CODE;
  return;
}

static unsigned long off_data = 0;
static unsigned long max_data = 0;
static int           stealth_init = BAD;

int hideout_hex_block(int fd, unsigned char * str, int len);
unsigned long first_hex_block(int fd, unsigned long * max);

/*
 * get hidden data from a block of hex data
 */
int sh_unix_getline_stealth  (int fd, char * str, int len)
{
  int add_off;

  /* initialize
   */
  if (stealth_init == BAD)
    {
      off_data = first_hex_block(fd, &max_data);
      if (off_data == 0)
	{
	  sh_error_handle (SH_ERR_FATAL, PLACE
			   NULL,
			   _("No data in stealth config file."));
	  exit (EXIT_FAILURE);
	}
      stealth_init = GOOD;
    }
  
  /* seek to proper position
   */
  lseek(fd, off_data, SEEK_SET);
     
  /* read one line
   */
  add_off = hideout_hex_block(fd, str, len);
  /*    fprintf(stdout, "%s", str); */
  off_data += add_off;


  if (0 != strncmp(str, _("[EOF]"), 5))
    return strlen(str);
  else
    return 0;

}

int hideout_hex_block(int fd, unsigned char * str, int len)
{

  register int  i, j, k;
  unsigned char c, e;
  register int  num;
  char          mask[9] = { 0x80, 0x40, 0x20, 0x10, 0x08, 0x04, 0x02, 0x01 };
  unsigned long here   = 0;
  unsigned long retval = 0;

  i = 0;
  while (i < len)
    {
      for (j = 0; j < 8; ++j)
	{

	  /* get a low byte, modify, read back */
	  for (k = 0; k < 2; ++k)
	    {
	      c = ' ';
	      do {
		do {
		  num = read (fd, &c, 1);
		} while (num == 0 && errno == EINTR);
		if (num == 0) return -1;
		++here; 
	      } while (c == '\n' || c == '\t' || c == '\r' || 
		       c == ' ');
	    }
	  

	  /* e is the value of the low byte
	   */
	  e = (unsigned char) sh_util_hexchar( c );
	  if ((e & mask[7]) != 0)  /* bit is set     */
	    str[i] |= mask[j];
	  else                     /* bit is not set */
	    str[i] &= ~mask[j];

	}
      if (str[i] == '\n') break;
      ++i;
    }
  if (i != 0)
    str[i] = '\0';
  else
    str[i+1] = '\0';
  retval += here;
  return retval;
}

/* get offset of first data block
 */
unsigned long first_hex_block(int fd, unsigned long * max)
{
  register int  i;
  register int  num = 1;
  char          c;
  int           nothex = 0;
  unsigned long retval = 0;
  register int  this_line = 0;
  char          theline[SH_BUFSIZE];

  *max = 0;

  while (1)
    {
      theline[0] = '\0';
      this_line  = 0;
      c          = '\0';
      while (c != '\n' && num > 0)
	{
	  do {
	    num = read (fd, &c, 1);
	  } while (num == 0 && errno == EINTR);
	  if (num > 0) theline[this_line] = c;
	  else         return 0;
	  this_line += num;
	}
      theline[this_line] = '\0';
      
      /* not only 'newline' */ 
      if (this_line > 60)
	{
	  nothex  = 0;
	  i       = 0;
	  while (nothex == 0 && i < (this_line-1))
	    {
	      if (! isxdigit(theline[i])) nothex = 1;
	      ++i;
	    }
	  if (nothex == 1) retval += this_line;
	}
      else
	{
	  nothex = 1;
	  retval += this_line;
	}

      if (nothex == 0)
	{
	  *max = 0; 
	  do {
	    do {
	      num = read (fd, theline, SH_BUFSIZE);
	    } while (num == 0 && errno == EINTR);
	    for (i = 0; i < num; ++i)
	      { 
		c = theline[i];
		if (c == '\n' || c == '\t' || c == '\r' || c == ' ') 
		  ;
		else if (!isxdigit(c))
		  break;
		else
		  *max += 1;
	      }
	  } while (num > 0);

	  *max /= 16;
	  return retval;
	}

    }
  return 0;
}

 /* ifdef SH_STEALTH */
#endif  
