/* SAMHAIN file system integrity testing                                   */
/* Copyright (C) 1999 Rainer Wichmann                                      */
/*                                                                         */
/*  This program is free software; you can redistribute it                 */
/*  and/or modify                                                          */
/*  it under the terms of the GNU General Public License as                */
/*  published by                                                           */
/*  the Free Software Foundation; either version 2 of the License, or      */
/*  (at your option) any later version.                                    */
/*                                                                         */
/*  This program is distributed in the hope that it will be useful,        */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/*  GNU General Public License for more details.                           */
/*                                                                         */
/*  You should have received a copy of the GNU General Public License      */
/*  along with this program; if not, write to the Free Software            */
/*  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.              */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>


#include "config_xor.h"

#ifdef SH_USE_UTMP

#ifdef HAVE_UTMPX_H

#include <utmpx.h>
#define SH_UTMP_S utmpx
#undef  ut_name
#define ut_name ut_user
#ifdef HAVE_UTXTIME
#undef  ut_time
#define ut_time        ut_xtime
#else
#undef  ut_time
#define ut_time        ut_tv.tv_sec
#endif

#else
#include <utmp.h>
#define SH_UTMP_S utmp
#endif


#ifdef HAVE_PATHS_H
#include <paths.h>
#endif

#if defined (SH_WITH_CLIENT) || defined (SH_STANDALONE) 


#include "samhain.h"
#include "sh_utils.h"
#include "sh_error.h"
#include "sh_modules.h"
#include "sh_utmp.h"


#ifdef TM_IN_SYS_TIME
#include <sys/time.h>
#else
#include <time.h>
#endif

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif 

#ifdef HAVE_DIRENT_H
#include <dirent.h>
#define NAMLEN(dirent) strlen((dirent)->d_name)
#else
#define dirent direct
#define NAMLEN(dirent) (dirent)->d_namlen
#ifdef HAVE_SYS_NDIR_H
#include <sys/ndir.h>
#endif
#ifdef HAVE_SYS_DIR_H
#include <sys/dir.h>
#endif
#ifdef HAVE_NDIR_H
#include <ndir.h>
#endif
#endif

#ifndef HAVE_LSTAT
#define lstat stat
#endif 

#ifndef UT_LINESIZE
#define UT_LINESIZE           12
#endif
#ifndef UT_NAMESIZE
#define UT_NAMESIZE            8
#endif
#ifndef UT_HOSTSIZE
#define UT_HOSTSIZE           16
#endif

#ifdef HAVE_UTMPX_H

#ifndef _PATH_UTMP
#ifdef   UTMPX_FILE
#define _PATH_UTMP   UTMPX_FILE
#else  
#error  You must define UTMP_FILE in the file config.h 
#endif
#endif
#ifndef _PATH_WTMP
#ifdef   WTMPX_FILE
#define _PATH_WTMP   WTMPX_FILE
#else
#error  You must define WTMP_FILE in the file config.h
#endif
#endif

#else

#ifndef _PATH_UTMP
#ifdef   UTMP_FILE
#define _PATH_UTMP   UTMP_FILE
#else  
#error  You must define UTMP_FILE in the file config.h 
#endif
#endif
#ifndef _PATH_WTMP
#ifdef   WTMP_FILE
#define _PATH_WTMP   WTMP_FILE
#else
#error  You must define WTMP_FILE in the file config.h
#endif
#endif

#endif

typedef struct log_user {
  char                ut_tty[UT_LINESIZE+1];    
  char                name[UT_NAMESIZE+1];
  char                ut_host[UT_HOSTSIZE+1]; 
  time_t              time; 
  struct log_user   * next;
} blah_utmp;

static char   terminated_line[UT_HOSTSIZE]; 

static struct SH_UTMP_S save_utmp;

static void sh_utmp_logout_morechecks(struct log_user   * user);
static void sh_utmp_login_morechecks(struct SH_UTMP_S * ut);
static void sh_utmp_addlogin (struct SH_UTMP_S * ut);
static void sh_utmp_check_internal(int mode);

int    ShUtmpLoginSolo    = 7;
int    ShUtmpLoginMulti   = 7;
int    ShUtmpLogout       = 7;
int    ShUtmpActive       = BAD;
time_t ShUtmpInterval     = 300;

sh_rconf sh_utmp_table[] = {
  {
    "SeverityLogin",
    sh_utmp_set_login_solo
  },
  {
    "SeverityLoginMulti",
    sh_utmp_set_login_multi
  },
  {
    "SeverityLogout",
    sh_utmp_set_logout_good
  },
  {
    "LoginCheckActive",
    sh_utmp_set_login_activate
  },
  {
    "LoginCheckInterval",
    sh_utmp_set_login_timer
  },
  {
    NULL,
    NULL
  },
};




#if defined (HAVE_SETUTENT) && defined (USE_SETUTENT)

#ifdef HAVE_UTMPX_H

#define sh_utmp_utmpname     utmpxname
#define sh_utmp_setutent     setutxent
#define sh_utmp_endutent     endutxent
#define sh_utmp_getutent     getutxent
#define sh_utmp_getutid      getutxid
#define sh_utmp_getutline    getutxline

#else

#define sh_utmp_utmpname     utmpname
#define sh_utmp_setutent     setutent
#define sh_utmp_endutent     endutent
#define sh_utmp_getutent     getutent
#define sh_utmp_getutid      getutid
#define sh_utmp_getutline    getutline

#endif

#else

/* BSD lacks getutent() etc.
 * utmpname(), setutent(), and endutent() return void,
 * so we do not perform much error handling.
 * Errors must be recognized by getutent() returning NULL.
 * Apparently, the application cannot check whether wtmp is empty,
 * or whether there was an fopen() error.
 */

static FILE * sh_utmpfile = NULL;
static char   sh_utmppath[80] = _PATH_UTMP;
static long   sh_utmp_feed_forward = 0;

static void sh_utmp_utmpname(const char * str)
{
  if (sh_utmpfile != NULL)
    {
      fclose (sh_utmpfile);
      sh_utmpfile = NULL;
    }

  strlcpy (sh_utmppath, str, 80);
  return;
}

static void sh_utmp_setutent()
{
  int error;

  if (sh_utmpfile == NULL) 
    {
      sh_utmpfile = fopen(sh_utmppath, "r");
      if (sh_utmpfile == NULL) 
	{
	  error = errno;
	  sh_error_handle (SH_ERR_ERR, PLACE
			   _("Error opening %s"),
			   _("Error opening %s: %s"),
			   sh_error_message(error));
	  return;
	}
    }
  (void) fseek (sh_utmpfile, 0L, SEEK_SET);
  (void) fseek (sh_utmpfile, sh_utmp_feed_forward, SEEK_CUR); 
  clearerr (sh_utmpfile);
  return;
}

static void sh_utmp_endutent()
{
  (void) fclose(sh_utmpfile);
  sh_utmpfile = NULL;
  return;
}

static struct SH_UTMP_S * sh_utmp_getutent()
{
  size_t in;
  static struct SH_UTMP_S out;
 
  if (sh_utmpfile == NULL)
    return NULL;
  in = fread (&out, sizeof(struct SH_UTMP_S), 1, sh_utmpfile);
  if (in != 1) {
    if (ferror (sh_utmpfile) != 0) {
      clearerr (sh_utmpfile);
      return NULL;
    } else {
      return NULL;
    }
  }
  return &out;
}

#ifdef USE_UNUSED

static struct SH_UTMP_S * sh_utmp_getutline(struct SH_UTMP_S * ut)
{
  struct SH_UTMP_S * out;
 
  while (1) {
      if ((out = sh_utmp_getutent()) == NULL) {
       	return NULL;
      }
#ifdef HAVE_UTTYPE  
      if (out->ut_type == USER_PROCESS || out->ut_type == LOGIN_PROCESS)
	if (strcmp(ut->ut_line, out->ut_line) == 0) 
	  return out;
#else
      if ( 0 != strncmp (out->ut_name, "reboot",   6) &&
	   0 != strncmp (out->ut_name, "shutdown", 8) &&
	   0 != strncmp (out->ut_name, "date",     4) )
	return out;
#endif
  }
  return NULL;
}

static struct SH_UTMP_S * sh_utmp_getutid(struct SH_UTMP_S * ut)
{
#ifdef HAVE_UTTYPE  
  struct SH_UTMP_S * out;

  if (ut->ut_type == RUN_LVL  || ut->ut_type == BOOT_TIME ||
      ut->ut_type == NEW_TIME || ut->ut_type == OLD_TIME) 
    {
      while (1) {
	if ((out = sh_utmp_getutent()) == NULL) {
	  return NULL;
	}
	if (out->ut_type == ut->ut_type) 
	  return out;
      }
    } 
  else if (ut->ut_type == INIT_PROCESS || ut->ut_type == LOGIN_PROCESS ||
	   ut->ut_type == USER_PROCESS || ut->ut_type == DEAD_PROCESS ) 
    {
      while (1) {
	if ((out = sh_utmp_getutent()) == NULL) {
	  return NULL;
	}
	if (strcmp(ut->ut_id, out->ut_id) == 0) 
	  return out;
      }
    }
#endif
  return NULL;
}
/* #ifdef USE_UNUSED */
#endif

/* #ifdef HAVE_SETUTENT */
#endif



static struct log_user   * userlist   = NULL;
static time_t  lastcheck;

/*************
 *
 * module init
 *
 *************/
int sh_utmp_init ()
{
  if (ShUtmpActive == BAD)
    return BAD;
  lastcheck  = time (NULL);
  userlist   = NULL;
  memset (&save_utmp, 0, sizeof(struct SH_UTMP_S));
  sh_utmp_check_internal (2); /* current logins */
  sh_utmp_check_internal (0);
  return GOOD;
}

/*************
 *
 * module cleanup
 *
 *************/
int sh_utmp_end ()
{
  struct log_user * user    = userlist;
  struct log_user * userold;
  while (user)
    {
      userold = user;
      user    = user->next;
      SH_FREE(userold);
    }
  return GOOD;
}


/*************
 *
 * module timer
 *
 *************/
int sh_utmp_timer (unsigned long tcurrent)
{
  if ((tcurrent - lastcheck) > ShUtmpInterval)
    {
      lastcheck  = tcurrent;
      return 1;
    }
  return 0;
}

/*************
 *
 * module check
 *
 *************/
int sh_utmp_check ()
{
  sh_error_handle (SH_ERR_INFO, PLACE
		   NULL, _("Checking logins."));
  sh_utmp_check_internal (1);
  return 0;
}

/*************
 *
 * module setup
 *
 *************/
/*
static void sh_utmp_set (int * flag, char * c)
{
  long val;

  val = strtol (c, (char **)NULL, 10);
  if (val < SH_ERR_ALL || val > SH_ERR_NOT)
    sh_error_handle (SH_ERR_ERR, PLACE
		     NULL, 
		     _("Utmp log level out of range, value clamped: %s"),
		     c);
  val = (val > SH_ERR_NOT ? SH_ERR_NOT : val);
  val = (val < SH_ERR_ALL ? SH_ERR_ALL : val);

  *flag = (int) val;
  return;
}
*/

int sh_utmp_set_login_solo  (char * c)
{
  char tmp[32];
  tmp[0] = '='; tmp[1] = '\0';
  strlcat (tmp, c, 32);
  sh_error_set_level (tmp, &ShUtmpLoginSolo);
  return 0;
}
int sh_utmp_set_login_multi (char * c)
{
  char tmp[32];
  tmp[0] = '='; tmp[1] = '\0';
  strlcat (tmp, c, 32);
  sh_error_set_level (tmp, &ShUtmpLoginMulti);
  return 0;
}
int sh_utmp_set_logout_good (char * c)
{
  char tmp[32];
  tmp[0] = '='; tmp[1] = '\0';
  strlcat (tmp, c, 32);
  sh_error_set_level (tmp, &ShUtmpLogout);
  return 0;
}

int sh_utmp_set_login_timer (char * c)
{
  long val;

  val = strtol (c, (char **)NULL, 10);
  if (val <= 0)
    sh_error_handle (SH_ERR_ERR, PLACE
		     NULL, 
		     _("Utmp timer <= NULL: %s"),
		     c);
  val = (val <= 0 ? 60 : val);

  ShUtmpInterval = (time_t) val;
  return 0;
}
int sh_utmp_set_login_activate (char * c)
{
  long val = -1;

  val = strtol (c, (char **)NULL, 10);
  if (val < 0 || val > 1)
    {

      sh_error_handle (SH_ERR_ERR, PLACE
		       NULL, 
		       _("Utmp activate: only 0|1 allowed: %s"),
		       c);
      return 1;
    }

  ShUtmpActive = (val == 0 ? BAD : GOOD);
  return 0;
}



/* for each login:
 *    - allocate a log record
 *    - link device.ut_record -> log_record
 *    - link user.ut_record   -> log_record
 */

static void sh_utmp_addlogin (struct SH_UTMP_S * ut)
{
  struct log_user   * user     = userlist;
  struct log_user   * username = userlist;
  struct log_user   * userold  = userlist;

  if (ut->ut_line[0] == '\0')
    return;

  /* for some stupid reason, AIX repeats the wtmp entry for logouts
   * with ssh
   */
  if (memcmp (&save_utmp, ut, sizeof(struct SH_UTMP_S)) == 0)
    {
      memset(&save_utmp, '\0', sizeof(struct SH_UTMP_S));
      return;
    }
  memcpy (&save_utmp, ut, sizeof(struct SH_UTMP_S));
      

  /* ------- find user -------- 
   */
  while (user != NULL) 
    {
      if (0 == strncmp(user->ut_tty, ut->ut_line, UT_LINESIZE) ) 
	break;
      userold = user;
      user = user->next;
    }

  while (username != NULL) 
    {
      if (0 == strncmp(username->name, ut->ut_name, UT_NAMESIZE) ) 
	break;
      username = username->next;
    }

  
#ifdef HAVE_UTTYPE  
  /* ---------- LOGIN -------------- */
  if (ut->ut_type == USER_PROCESS) 
    {
      if (user == NULL)
	{
	  user = SH_ALLOC(sizeof(struct log_user));
	  user->next       = userlist;
	  userlist         = user;
	}
      strlcpy(user->ut_tty,  ut->ut_line, UT_LINESIZE+1);
      strlcpy(user->name,    ut->ut_name, UT_NAMESIZE+1);
#ifdef HAVE_UTHOST
      strlcpy(user->ut_host, ut->ut_host, UT_HOSTSIZE+1);
#endif
      user->time = ut->ut_time;


      if (username == NULL                              /* not yet logged in */
	  || 0 == strncmp(ut->ut_line, _("ttyp"), 4)    /* in virt. console  */
	  || 0 == strncmp(ut->ut_line, _("ttyq"), 4)    /* in virt. console  */
	  ) {
	sh_error_handle( ShUtmpLoginSolo, PLACE
			 NULL,
#ifdef HAVE_UTHOST
			 _("Login -- %8s  %12s  %16s  %s"),
#else
			 _("Login -- %8s  %12s  %s"),
#endif
			 user->name,
			 user->ut_tty,
#ifdef HAVE_UTHOST
			 user->ut_host,
#endif
			 sh_unix_time (user->time)
			 );
      } else
	if (0 != strncmp(ut->ut_line, _("ttyp"), 4) &&
	    0 != strncmp(ut->ut_line, _("ttyq"), 4))
	{       
	sh_error_handle( ShUtmpLoginMulti, PLACE
			 NULL,
#ifdef HAVE_UTHOST
			 _("Multiple login -- %8s  %12s  %16s  %s"),
#else
			 _("Multiple login -- %8s  %12s  %s"),
#endif
			 user->name,
			 user->ut_tty,
#ifdef HAVE_UTHOST
			 user->ut_host,
#endif
			 sh_unix_time (user->time)
			 );
      }
      
      sh_utmp_login_morechecks(ut);
      return;
    }


  /* ---------  LOGOUT ---------------- */
  else if (ut->ut_name[0] == '\0'
	   || ut->ut_type == DEAD_PROCESS  /* solaris does not clear ut_name */
	   )
    {
      if (user != NULL)
	{
	  sh_error_handle( ShUtmpLogout, PLACE
			   NULL,
#ifdef HAVE_UTHOST
			   _("Logout -- %8s  %12s  %16s  %s"),
#else
			   _("Logout -- %8s  %12s  %s"),
#endif
			   user->name,
			   user->ut_tty,
#ifdef HAVE_UTHOST
			   user->ut_host,
#endif
			   sh_unix_time (user->time)
			   );
	  userold->next = user->next;
	  if (user == userlist)
	    userlist = user->next;
	  sh_utmp_logout_morechecks(user);
	  SH_FREE(user);
	  user = NULL;
	}
      else
	{
	  strlcpy(terminated_line, ut->ut_line, UT_HOSTSIZE);
	  sh_error_handle( ShUtmpLogout, PLACE
			   NULL,
			   _("Logout -- %12s  %s"),
			   terminated_line,
			   sh_unix_time (ut->ut_time)
			   );
	}
      return;
    }

  /* default */
  return;

  /* #ifdef HAVE_UTTYPE                   */
#else
  if (user == NULL)   /* probably a login */
    {
      user = SH_ALLOC(sizeof(struct log_user));
      strlcpy(user->ut_tty,  ut->ut_line, UT_LINESIZE+1);
      strlcpy(user->name,    ut->ut_name, UT_NAMESIZE+1);
#ifdef HAVE_UTHOST
      strlcpy(user->ut_host, ut->ut_host, UT_HOSTSIZE+1);
#endif
      user->time       = ut->ut_time;
      user->next       = userlist;
      userlist         = user;

      sh_error_handle( ShUtmpLoginSolo, PLACE
		       NULL,
#ifdef HAVE_UTHOST
		       _("Login -- %8s  %12s  %16s  %s"),
#else
		       _("Login -- %8s  %12s  %s"),
#endif
		       user->name,
		       user->ut_tty,
#ifdef HAVE_UTHOST
		       user->ut_host,
#endif
		       sh_unix_time (user->time)
		       );
      sh_utmp_login_morechecks(ut);
    }
  else  /* probably a logout */
    {
      sh_error_handle( ShUtmpLogout, PLACE
		       NULL,
#ifdef HAVE_UTHOST
		       _("Logout --  %8s  %12s  %16s  %s"),
#else
		       _("Logout --  %8s  %12s %s"),
#endif
		       user->name,
		       user->ut_tty,
#ifdef HAVE_UTHOST
		       user->ut_host,
#endif
		       sh_unix_time (user->time)
		       );
      sh_utmp_logout_morechecks(user);
      userold->next = user->next;
      SH_FREE(user);
      user = NULL;
    }
#endif

  return;
}

static time_t        lastmod  = 0;
static off_t         lastsize = 0;
static unsigned long lastread = 0;

char * mode_path[] = { _PATH_WTMP, _PATH_WTMP, _PATH_UTMP };
    
static void sh_utmp_check_internal (int mode)
{
  struct stat   buf;
  int           error;
  struct SH_UTMP_S * ut;
  unsigned long this_read = 0;

  /* error if no access
   */
  if (0 != lstat(mode_path[mode], &buf)) 
    {
      error = errno;
      sh_error_handle(SH_ERR_ERR, PLACE
		      _("Cannot access %s"),
		      _("Failed lstat() on %s: %s"),
		      mode_path[mode],
		      sh_error_message(error));
      return;
    }

  /* modification time
   */
  if (mode < 2)
    {
      if (buf.st_mtime <= lastmod) 
	return;
      else
	lastmod = buf.st_mtime;
    }

  /* file size
   */
  if (buf.st_size < lastsize && mode < 2) 
    { 
      sh_error_handle(SH_ERR_STAMP, PLACE
		      NULL,
		      _("Logfile rotation for %s occured."),
		      mode_path[mode]);
      lastread = 0;
#ifndef USE_SETUTENT
      sh_utmp_feed_forward = 0L;
#endif
    }

  if (mode < 2)
    lastsize = buf.st_size;

  if (buf.st_size == 0) 
    return;

  sh_utmp_utmpname(mode_path[mode]);
  sh_utmp_setutent();

  /* 
   * feed forward if initializing
   * we need to do this here
   */
  if (mode < 2)
    {
      while (this_read < lastread) {
	ut = sh_utmp_getutent();
	++this_read;
      }
    }

  /* start reading
   */
  this_read = 0;
  while (1) {
    ut = sh_utmp_getutent();
    if (ut == NULL) 
      break;
    if (mode == 1 || (mode == 2 && ut->ut_user[0] != '\0'
#ifdef HAVE_UTTYPE
		      && ut->ut_type != DEAD_PROCESS
#endif
		      ))
      sh_utmp_addlogin (ut);
    /*************************************************
    printf("%8s | %10s | %10s | %d %d | %16s | %ld\n", 
	   ut->ut_user, ut->ut_id, ut->ut_line, 
	   (int) ut->ut_type, (int) ut->ut_pid, 
	   ut->ut_host, ut->ut_time);
    ***************************************************/
    ++this_read;
  }

  sh_utmp_endutent();

  if (mode < 2)
    {
      lastread += this_read;
#ifndef USE_SETUTENT
      sh_utmp_feed_forward += (long) (this_read * sizeof(struct SH_UTMP_S));
      lastread = 0;
#endif
    }

  return;
}


static void sh_utmp_login_morechecks(struct SH_UTMP_S * ut)
{
  return;
}

static void sh_utmp_logout_morechecks(struct log_user   * user)
{
  return;
}

#endif


/* #ifdef SH_USE_UTMP */
#endif



