/* LINTLIBRARY */
/*
 * This is the file with all the library routines in it
 *
 * Author information:
 * Matt Bishop
 * Department of Computer Science
 * University of California at Davis
 * Davis, CA  95616-8562
 * phone (916) 752-8060
 * email bishop@cs.ucdavis.edu
 *
 * This code is placed in the public domain.  I do ask that
 * you keep my name associated with it, that you not represent
 * it as written by you, and that you preserve these comments.
 * This software is provided "as is" and without any guarantees
 * of any sort.
 *
 * Compilation notes:
 * * this does NOT use malloc(3), but fixed storage.  this means we
 *   do lots of bounds checking, but it still is faster, and smaller,
 *   than forcing inclusion of malloc.  All buffers etc. are of size
 *   MAXFILENAME (defined in trustfile.h); to get more room, recompile
 *   with this set larger.
 * * if you support the following directory semantics, define STICKY;
 *   otherwise, undefine it
 *	"if a directory is both world-writeable AND has the sticky bit
 *	 set, then ONLY the owner of an existing file may delete it"
 *   On some systems (eg, IRIX), you can delete the file under these
 *   conditions if the file is world writeable.  Foor our purposes,
 *   this is irrelevant since if the file is world-writeable it is
 *   untrustworthy; either it can be replaced with another file (the
 *   IRIX version) or it can be altered (all versions).
 *   if this is true and STICKY is not set, the sticky bit is ignored
 *   and the directory will be flagged as untrustworthy, even when only
 *   a trusted user could delete the file
 * * this uses a library call to get the name of the current working
 *   directory.  Define the following to get the various versions:
 *   GETCWD	for Solaris 2.x, SunOS 4.1.x, IRIX 5.x
 *			char *getcwd(char *buf, int bufsz);
 *		where buf is a buffer for the path name, and bufsz is
 *		the size of the buffer; if the size if too small, you
 *		get an error return (NULL)
 *   GETWD	for Ultrix 4.4
 *			char *getwd(char *buf)
 *		where buf is a buffer for the path name, and it is
 *		assumed to be at lease as big as MAXPATHLEN.
 *		*** IMPORTANT NOTE ***
 *		Ultrix supports getcwd as well, but it uses popen to
 *		run the command "pwd" (according to the manual).  This
 *		means it's vulnerable to a number of attacks if used
 *		in a privileged program.  YOU DON'T WANT THIS.
 * * the debugging flag DEBUG prints out each step of the file name
 *   checking, as well as info on symbolic links (if S_IFLNK defined),
 *   file name canonicalization, and user, group, and permission for
 *   each file or directory; this is useful if you want to be sure
 *   you're checking the right file
 *
 * Version information:
 * 1.0		December 28, 1995	Matt Bishop
 */
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <grp.h>
#include <pwd.h>
#include <string.h>
#include <unistd.h>

#include "trustfile.h"

/*
 * the get current working directory function
 * every version of UNIX seems to have its own
 * idea of how to do this, so we group them by
 * arguments ...
 * all must return a pointer to the right name
 */

#include "config_xor.h"

#if defined(HAVE_GETCWD) && !defined(HAVE_BROKEN_GETCWD)
#define GETCWD
#else
#ifdef HAVE_GETWD
#define GETWD
#endif
#endif

#ifdef GETCWD
#define CURDIR(buf,nbuf)	getcwd((buf), (nbuf))
#endif
#ifdef GETWD
#define CURDIR(buf,nbuf)	getwd((buf))
#endif

/*
 * this checks to see if there are symbolic links
 * assumes the link bit in the protection mask is called S_IFLNK
 * (seems to be true on all UNIXes with them)
 */
#ifndef S_IFLNK
#	define	lstat	stat	/* no such call as lstat (follow link) */
#endif


/*
 * these are useful global variables
 *
 * first set: who you gonna trust, by default?
 * 	if the user does not specify a trusted or untrusted set of users,
 *	all users are considered untrusted EXCEPT:
 *	UID 0 -- root	as root can do anything on most UNIX systems, this
 *			seems reasonable
 *	tf_euid -- programmer-selectable UID
 *			if the caller specifies a specific UID by putting
 *			it in this variable, it will be trusted; this is
 *			typically used to trust the effective UID of the
 *			process (note: NOT the real UID, which will cause all
 *			sorts of problems!)  By default, this is set to -1,
 *			so if it's not set, root is the only trusted user
 *	NOTE: EUIDSLOT is the index in rootonly where the EUID (tf_euid)
 *	goes.  BE SURE THIS IS SET TO THE NEXT-TO-LAST ELEMENT OR YOU
 *	WON'T GET THE RIGHT RESULTS!!
 */
/* static int rootonly[] = { 0, -1, -1 };        the default trust list */
/* int tf_euid = -1;			      space for EUID of process */
/* #define EUIDSLOT	1		  location in rootonly for EUID */

/* modified Tue Feb 22 10:36:44 NFT 2000 Rainer Wichmann                */
static int test_rootonly[] = { ALWAYS_TRUSTED };

int rootonly[] = { ALWAYS_TRUSTED, -1, -1, -1, -1, -1, -1, -1, -1, -1 };
int tf_euid = -1;
/* #define EUIDSLOT_PRE	1 */
int EUIDSLOT = sizeof(test_rootonly)/sizeof(int);



/*
 * second set: how do you report problems?
 *	tf_errno	on return when an error has occurred, this is set
 *			to the code indicating the reason for the error:
 *			   TF_BADFILE	passed NULL for pointer to file name
 *			   TF_BADNAME	could not expand to full path name
 *			   TF_BADSTAT	stat failed; usu. file doesn't exist
 *			   TF_BADUID	owner untrusted
 *			   TF_BADGID	group untrusted & can write
 *			   TF_BADOTH	anyone can write
 *			the value is preserved across calls where no error
 *			occurs, just like errno(2)
 *	tf_path		if error occurs and a file name is involved, this
 *			contains the file name causing the problem
 */
int tf_errno;				/* error code for trust function */
char tf_path[MAXFILENAME];		/* error path for trust function */

/*
 * entry point:
 *	void dirz(char *path)
 * purpose:
 * 	reduce the path name by applying these rules:
 * 	(1) "/./" -> "/" 	(2) "/xxx/../" -> "/"
 * 	(3) "//" -> "/"
 * arguments:
 *	char *path	name to be cleaned up
 * return values:
 *	path		contains the cleaned-up name; note that
 *			this routine never adds chars, so you
 *			won't have overflow problems.  if it
 *			is NULL going in, it's NULL coming out
 * calls:
 *	strcpy		string copy; note we copy a trailing part
 *			of the array to a preceding position, so
 *			we can't have buffer overflow
 * gotchas:
 *	* heaven help you if the path isn't a full path name, because
 *	  this sucker won't; in patricular, "../" will give you the
 *	  wrong answer!
 */
#ifdef __STDC__
static void dirz(char *path)
#else
static void dirz(path)
char *path;		/* path name to be cleaned up */
#endif
{
	register char *p = path;/* points to rest of path to clean up */
	register char *q;	/* temp pointer for skipping over stuff */
        char * swp;

	/*
	 * standard error checking
	 */
	if (path == NULL)
		return;

	swp = (char *) malloc (strlen (path) + 1);
	if (!swp)
	  {
	    fprintf (stderr, "trustfile: dirz: Out of memory (%ld bytes)\n",
		     (long)(strlen(path)+1));
	    _exit (EXIT_FAILURE);
	  }

	/*
	 * loop over the path name until everything is checked
	 */
	while(*p){
		/* not one of 3 rules -- skip */
		if (*p != '/'){
			p++;
			continue;
		}
		/* is it "/./" or "/."? */
		if (p[1] == '.' && (p[2] == '/' || !p[2])){
			/* yes -- delete "/." */
		        (void) strcpy(swp, &p[2]);
			(void) strcpy(p, swp);
			/* special case "/." as full path name */
			/* this is "/", not "" 		       */
			if (p == path && !*p){
				*p++ = '/';
				*p = '\0';
			}
		}
		/* is it "//" (multiple /es)? */
		else if (p[1] == '/'){
			/* yes -- skip them */
			for(q = &p[2]; *q == '/'; q++)
				;
                        (void) strcpy(swp, q);
			(void) strcpy(&p[1], swp);
		}
		/* is it "/../" or "/.."? */
		else if (p[1] == '.' && p[2] == '.' && (p[3] == '/' || !p[3])){
			/* yes -- if it's root, delete .. only */
			if (p == path){
			        (void) strcpy(swp, &p[3]);
				(void) strcpy(p, swp);
			}
			else{
				/* nope -- back up over previous component */
				q = p - 1;
				while(q != path && *q != '/')
					q--;
				/* now wipe it out */
                                (void) strcpy(swp, &p[3]);
				(void) strcpy(q, swp);
				p = q;
			}
		}
		else
			p++;
	}
	free (swp);
	return;
}
			

/*
 * entry points:
 *	char *getfname(char *fname, char *rbuf, int rsz)
 * purpose:
 *	canonicalize a file name
 * arguments:
 *	char *fname	name to be canonicalized
 *	char *rbuf	space for canonical name (must be
 *			allocated before call)
 *	int rsz		number of bytes in rbuf
 * return values:
 *	rbuf		success; canonical name in rbuf
 *	NULL		failure; means an invalid argument,
 *			or the get current working directory
 *			call failed (note no error is printed)
 * calls:
 *	dirz		collapse multiple /es, .s, and ..s in
 *			a path name (in this file)
 *	getcwd		get current working directory
 * gotchas:
 *	* if rbuf is too small, the path name will be wrong.
 *	  the caller must ensure it's big enough.
 *	* assumes a normal UNIX file system:
 *	  a/./b is a/b
 *	  a///b is a/b
 *	  a/x/../b is a/b
 *	* heaven help you if your cwd disappears during this thing ...
 */

/* Jan 24, 2000 Rainer Wichmann:
 * the doc on strcpy says: "The strings may not overlap, and ...",
 * therefore all strcpy's replaced by pairs of strcpy's with a temporary
 * swap array
 */

#ifdef __STDC__
static char *getfname(char *fname, char *rbuf, int rsz)
#else
static char *getfname(fname, rbuf, rsz)
char *fname;			/* file name to be canonicalized */
char *rbuf;			/* where to put canonicalization */
int rsz;			/* space for canonicalization */
#endif
{
	register char *p = rbuf;	/* used to load array */

	/*
	 * do the initial checking
	 * NULL pointer
	 */
	if (fname == NULL || rbuf == NULL || rsz <= 0)
		return(NULL);

#ifdef DEBUG
	printf("%s CANONICALIZES TO ", fname);
#endif

	/* already a full path name */
	if (*fname == '/')
		rbuf[0] = '\0';
	else{
		/*
		 * it's a relative path name
		 * now get the full path name
		 * of the current working directory
		 */
		if (CURDIR(rbuf, rsz)  == NULL){
#ifdef DEBUG
			printf("can't get current working directory\n");
#endif
			tf_errno = TF_BADNAME;
			return(NULL);
		}
	}

	/*
	 * append the file name and reduce
	 */
	if (fname != NULL && *fname != '\0'){
		/* we can't use strcat since */
		/* it does not check bounds  */
		for(p = rbuf; *p; p++)
			;
		if (p != rbuf && p < rbuf + rsz)
			*p++ = '/';
		while(*fname && p < rbuf + rsz)
			*p++ = *fname++;
		/* if too little room, error */
		if (p == rbuf + rsz){
#ifdef DEBUG
			printf("not enough room\n");
#endif
			tf_errno = TF_NOROOM;
			return(NULL);
		}
		*p = '\0';
	}
	dirz(rbuf);

	/*
	 * return it
	 */
#ifdef DEBUG
	printf("%s\n", rbuf);
#endif
	return(rbuf);
}

/*
 * entry point:
 *	int isin(int n, int *list)
 * purpose:
 * 	scan to see if the integer n is in the list of integers pointed to
 * 	by list (which is terminated by -1)
 * arguments:
 *	int n 		number to be looked for
 *	int *list	list to be searched (-1 terminates it)
 * return values:
 *	1		yes, it's in there
 *	0		no, it's not in there, or list is NULL
 * calls:
 *	nothing
 * gotchas:
 *	* you're SOL if -1 is a valid integer to be foud
 */
#ifdef __STDC__
static int isin(int n, int *list)
#else
static int isin(n, list)
int n;				/* number to look for */
int *list;			/* where to look */
#endif
{
	/*
	 * error check -- if no list, n's not in it ...
	 */
	if (list == NULL)
		return(0);
	/*
	 * just walk the list until it's done or you find it
	 */
	while(*list != -1 && *list != n)
		list++;
	/*
	 * say what happened
	 */
	return(*list == n);
}

/*
 * entry point:
 *	int isingrp(int grp, int *ulist)
 * purpose:
 * 	scan ulist to see if any UID in that list is a member of the group grp
 * 	UID list is terminated by -1
 * arguments:
 *	int grp 	ID of group to be checked
 *	int *ulist	set of UIDs to be checked against group members
 * return values:
 *	1		yes, one of the UIDs in ulist is a member of the
 *			group with GID grp
 *	0		no, none are, or there is no such group
 * calls:
 *	getpwnam	maps user name to user info (such as UID)
 *	getgrgid	maps group name to group info (such as list
 *			of users making up the group)
 */
#ifdef __STDC__
static int isingrp(int grp, int *ulist)
#else
static int isingrp(grp, ulist)
int grp;			/* group number */
int *ulist;			/* list of UIDs */
#endif
{
	register struct passwd *w;	/* info about group member */
	register int *u;		/* points to current ulist member */
	register char **p;		/* points to current group member */
	register struct group *g;	/* pointer to group information */

	/*
	 * get group information
	 */
	if ((g = getgrgid(grp)) == NULL)
		return(0);

	/*
	 * see if any member of the list ulist
	 * is a member of the group grp
	 */
	for(p = g->gr_mem; *p != NULL; p++)
		for(u = ulist; *u != -1; u++)
			/* map user name to UID and compare */
			if ((w = getpwnam(*p)) != NULL && *u == w->pw_uid)
				return(1);
	
	/*
	 * nope, none in the group
	 */
	return(0);
}

/*
 * entry points:
 *	int trustfile(char *fname, int *okusers, int *badusers)
 * purpose:
 *	determine if a file can be "trusted".
 *	Define "trust" as meaning that the file can be altered by
 *	only a trusted subset of users.  Then, if the file OR any
 *	of its ancestor directories can be altered by any other
 *	user, the file is "untrustworthy".  Otherwise, it's trustworthy.
 *	A "trusted user" is defined either as a user in a set of
 *	trusted users, or as a user not in a set of untrusted users.
 *	Note these definitions are mutually exclusive.
 * arguments:
 *	char *fname	name of file to be checked
 *	int *okusers	pointer to a -1 terminated list of UIDs
 *			belonging to trusted users
 *	int *badusers	pointer to a -1 terminated list of UIDs
 *			belonging to untrusted users
 * return values:
 *	TF_YES		file is trustworthy
 *	TF_NO		file is untrustworthy; tf_path contain
 *			the file/directory which is untrustworthy,
 *			and tf_errno says why
 *	TF_ERROR	error occurred, so can't say; tf_path contains
 *			the file/directory which caused the problem
 *			(if any) and tf_errno says what the problem
 *			was (caller may need to go to errno for more
 *			info)
 * 			*** (More details on the codes in trustfile.h)
 * calls:
 *	getfname	expand file name into full path name
 *	lstat/stat	get owner, group, and protection mask
 *			of file (lstat if system supports symbolic
 *			links -- S_IFLNK defined -- else stat)
 *	readlink	get what symbolic link points to (only
 *			if system supports symbolic links --
 *			S_IFLNK defined)
 *	trustfile	recursive if symbolic link involved (ie,
 *			S_I(FLNK defined)
 *	isin		see if a number (UID) is in a list
 *			(trusted/untrusted users)
 *	isingrp		see if any member of a list is in a given
 *			group
 *	printf		used for debugging (if DEBUG set)
 * algorithm:
 * 1	if okusers is non-NULL, badusers is ignored and only users
 *	named in okusers are trusted (note: if root isn't there, the
 *	superuser is untrusted)
 * 2	if okusers is NULL, badusers is used; then all users EXCEPT
 *	those named in badusers are trusted
 * 3	canonicalize the file name, since we need to check from the
 * 	root on down
 * 4	get UID, GID, mode info about the file
 * 5	if a symlink, resolve the link to a full path name; this is
 *	an alternate way into the file, so it too must be checked.
 *	when done, continue from here
 * 6	check if owner is trusted; if not, object is untrusted, so stop.
 * 7	check if file is group writeable
 * 8	if so, check that all members of the group are trusted; if not, stop.
 * 9	check that world cannot write to the file; if they can, stop
 * A	if anything's left in the path name, tack on next component
 *	and go to 4
 */
#ifdef __STDC__
int trustfile(char *fname, int *okusers, int *badusers)
#else
int trustfile(fname, okusers, badusers)
char *fname;			/* name of file to be checked */
int *okusers;			/* list of trusted users */
int *badusers;			/* list of untrusted users */
#endif
{
	char fexp[MAXFILENAME];	/* file name fully expanded */
	register char *p = fexp;/* used to hold name to be checked */
	struct stat stbuf;	/* used to check file permissions */
	char c;			/* used to hold temp char */

	/*
	 * sanity checks -- first, be sure there is a path
	 */
	if (fname == NULL){
		tf_errno = TF_BADFILE;
		return(TF_ERROR);
	}
	/*
	 * nex expand to the full file name
	 * getfname sets tf_errno as appropriate
	 */
	if (getfname(fname, fexp, MAXFILENAME) == NULL)
		return(TF_ERROR);
	/*
	 * if no trust extended, use defaults
	 * that's root and the current effective UID
	 */
	if (okusers == NULL && badusers == NULL){
		okusers = rootonly;
		rootonly[EUIDSLOT] = tf_euid;
	}

	/*
	 * now loop through the path a component at a time
	 * note we have to special-case root
	 */
	while(*p){
		/*
		 * get next component
		 */
		while(*p && *p != '/')
			p++;
		/* save where you are */
		if (p == fexp){
			/* keep the / if it's the root dir */
			c = p[1];
			p[1] = '\0';
		}
		else{
			/* clobber the / if it isn't the root dir */
			c = *p;
			*p = '\0';
		}

#ifdef DEBUG
		/*
		 * say what you're looking at
		 */
		printf("%s: ", fexp);
#endif
		/*
		 * now get the information
		 */
		if (lstat(fexp, &stbuf) < 0){
#ifdef DEBUG
			/* oops ... */
			printf("cannot stat\n");
#endif
			(void) strcpy(tf_path, fexp);
			tf_errno = TF_BADSTAT;
			return(TF_ERROR);
		}
#ifdef DEBUG
		/*
		 * print out owner group, file protection modes
		 */
		printf("UID=%d GID=%d mask=%06o\n",
			(int) stbuf.st_uid, (int) stbuf.st_gid,
				(unsigned) stbuf.st_mode);
#endif
#ifdef S_IFLNK
		/* 
		 * if it's a symbolic link, recurse
		 */
		if ((stbuf.st_mode&S_IFLNK) == S_IFLNK){
			/*
			 * this is tricky
			 * if the symlink is to an absolute path
			 * name, just call trustfile on it; but
			 * if it's a relative path name, it's 
			 * interpreted WRT the current working
			 * directory AND NOT THE FILE NAME!!!!!
			 * so, we simply put /../ at the end of
			 * the file name, then append the symlink
			 * contents; trustfile will canonicalize
			 * this, and the /../ we added "undoes"
			 * the name of the symlink to put us in
			 * the current working directory, at
			 * which point the symlink contents (appended
			 * to the CWD) are interpreted correctly.
			 * got it?
			 */
			char csym[MAXFILENAME];	/* contents of symlink file */
			char full[MAXFILENAME];	/* "full" name of symlink */
			register char *b, *t;	/* used to copy stuff around */
			register int lsym;	/* num chars in symlink ref */
			register int i;		/* trustworthy or not? */

			/*
			 * get what the symbolic link points to
			 */
			lsym = readlink(fexp, csym, MAXFILENAME);
			csym[lsym] = '\0';
#ifdef DEBUG
			/*
			 * announce this
			 */
			printf("%s: SYMLINK TO %s\n", fexp, csym);
#endif
			/*
			 * relative or absolute referent?
			 */
			if (csym[0] != '/'){
				/* initialize pointers */
				b = full;
				/* copy in base path */
				t = fexp;
				while(*t && b < &full[MAXFILENAME])
					*b++ = *t++;
				/* smack on the /../ */
				t = "/../";
				while(*t && b < &full[MAXFILENAME])
					*b++ = *t++;
				/* append the symlink referent */
				t = csym;
				while(*t && b < &full[MAXFILENAME])
					*b++ = *t++;
				/* see if we're too big */
				if (*t || b == &full[MAXFILENAME]){
					/* yes -- error */
					tf_errno = TF_NOROOM;
					(void) strcpy(tf_path, fexp);
					return(TF_ERROR);
				}
				/* nope */
				*b = '\0';
			}
			else{
				/* absolute -- just copy */
				/* overflow can't occur as the arrays */
				/* are the same size		      */
				(void) strcpy(full, csym);
			}
			/*
			 * now check out this file and its ancestors
			 */
			if ((i = trustfile(full, okusers, badusers)) != TF_YES)
				return(i);
			/*
			 * okay, this part is valid ... let's check the rest
		 	 * put the / back
			 */
			if (p == fexp){
				/* special case for root */
				p[1] = c;
				p++;
			}
			else{
				/* ordinary case for everything else */
				*p = c;
				if (*p)
					p++;
			}
			continue;
		}
#endif

			
		/*
		 * if the owner is not trusted then -- as the owner can
		 * change protection modes -- he/she can write to the
		 * file regardless of permissions, so bomb
		 */
		if (((okusers != NULL && !isin((int)stbuf.st_uid, okusers)) ||
		     (badusers != NULL && isin((int)stbuf.st_uid, badusers)))){
			tf_errno = TF_BADUID;
			(void) strcpy(tf_path, fexp);
			return(TF_NO);
		}
		/*
		 * if a group member can write but the
		 * member is not trusted, bomb; but if
		 * sticky bit semantics are honored, it's
		 * okay
		 */
		if (((stbuf.st_mode&S_IWGRP) == S_IWGRP) &&
		   ((okusers != NULL && !isingrp((int)stbuf.st_gid, okusers)) ||
		   (badusers != NULL && isingrp((int)stbuf.st_gid, badusers)))
#ifdef STICKY
			&& ((stbuf.st_mode&S_IFDIR) != S_IFDIR ||
				(stbuf.st_mode&S_ISVTX) != S_ISVTX)
#endif
		){
			tf_errno = TF_BADGID;
			(void) strcpy(tf_path, fexp);
			return(TF_NO);
		}
		/*
		 * if other can write, bomb; but if the sticky
		 * bit semantics are honored, it's okay
		 */
		if (((stbuf.st_mode&S_IWOTH) == S_IWOTH)
#ifdef STICKY
			&& ((stbuf.st_mode&S_IFDIR) != S_IFDIR ||
				(stbuf.st_mode&S_ISVTX) != S_ISVTX)
#endif
		){
			tf_errno = TF_BADOTH;
			(void) strcpy(tf_path, fexp);
			return(TF_NO);
		}
		/*
		 * put the / back
		 */
		if (p == fexp){
			/* special case for root */
			p[1] = c;
			p++;
		}
		else{
			/* ordinary case for everything else */
			*p = c;
			if (*p)
				p++;
		}
	}
	/*
	 * yes, it can be trusted
	 */
	(void) strcpy(tf_path, fexp);
	return(TF_YES);
}

